# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


from math import log10
import warnings

import numpy as np

from magnopy._data_validation import _validated_units
from magnopy._constants._units import _ENERGY_UNITS


# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")

_C1 = 1e-4
_C2 = 0.9


def _cubic_interpolation(alpha_l, alpha_h, phi_l, phi_h, der_l, der_h):
    r"""
    Computes the minimum of a cubic interpolation for the function f(alpha) with
    values f_l, f_h and derivatives fp_l, fp_h at the points alpha_l, alpha_h.

    Parameters
    ----------
    alpha_l : float
        Lower bound of the interval.
    alpha_h : float
        Upper bound of the interval.
    phi_l : float
        Value of the function at alpha_l.
    phi_h : float
        Value of the function at alpha_h.
    der_l : float
        Derivative of the function at alpha_l.
    der_h : float
        Derivative of the function at alpha_h.

    Returns
    -------
    alpha_min : float
        Position of the minimum of the cubic interpolation for the function f(alpha).
    """

    if abs(alpha_l - alpha_h) < np.finfo(float).eps:
        return alpha_l

    d_1 = der_l + der_h - 3 * (phi_l - phi_h) / (alpha_l - alpha_h)

    if d_1**2 - der_l * der_h < 0:
        if phi_l <= phi_h:
            return alpha_l
        else:
            return alpha_h

    d_2 = np.sign(alpha_h - alpha_l) * np.sqrt(d_1**2 - der_l * der_h)

    return alpha_h - (alpha_h - alpha_l) * (der_h + d_2 - d_1) / (
        der_h - der_l + 2 * d_2
    )


def _rotate_sd(reference_sd, rotation):
    r"""

    Parameters
    ----------
    reference_sd : (M, 3) :numpy:`ndarray`
        Reference direction of the spin vectors.
    rotation : (M*3,) :numpy:`ndarray`
        Rotation of the spin vectors parameterized with the skew-symmetric matrix.

    Returns
    -------
    directions : (I, 3) :numpy:`ndarray`
        Rotated set of direction vectors.
    """

    directions = reference_sd.copy()

    ax = rotation[::3]
    ay = rotation[1::3]
    az = rotation[2::3]

    thetas = np.sqrt(ax**2 + ay**2 + az**2)

    r = []
    for alpha in range(len(thetas)):
        theta = thetas[alpha]

        if theta < np.finfo(float).eps:
            continue

        r = np.array([ax[alpha], ay[alpha], az[alpha]]) / theta

        directions[alpha] = (
            np.cos(theta) * directions[alpha]
            + np.sin(theta) * np.cross(r, directions[alpha])
            + (1 - np.cos(theta)) * r * (r @ directions[alpha])
        )

    return directions


class Energy:
    r"""
    Classical energy of the spin Hamiltonian.

    This class is optimized for calculation of the energy for any spin directions for the
    given Hamiltonian.

    .. important::
        If the spin Hamiltonian is modified, then a new instance of the energy class
        should be created.

    Parameters
    ----------

    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian for the calculation of energy.

    Examples
    --------

    First, one need to create some spin Hamiltonian

    .. doctest::

        >>> import numpy as np
        >>> import magnopy
        >>> cell = np.eye(3)
        >>> atoms = dict(
        ...     names=["Fe"], spins=[1.5], g_factors=[2], positions=[[0, 0, 0]]
        ... )
        >>> convention = magnopy.Convention(
        ...     multiple_counting=True, spin_normalized=False, c21=1, c22=-1
        ... )
        >>> spinham = magnopy.SpinHamiltonian(
        ...     cell=cell, atoms=atoms, convention=convention
        ... )

    Then, add some parameters to the Hamiltonian

    .. doctest::

        >>> spinham.add_21(alpha=0, parameter=np.diag([0, 0, -1]))
        >>> spinham.add_22(
        ...     alpha=0,
        ...     beta=0,
        ...     nu=(1, 0, 0),
        ...     parameter=magnopy.converter22.from_iso(iso=1),
        ... )

    Now everything is ready to create an instance of the Energy class

    .. doctest::

        >>> energy = magnopy.Energy(spinham)

    Finally, ``energy`` can be used to compute classical energy of the Hamiltonian,
    its gradient, torque or search for the local minima.

    .. doctest::

        >>> sd1 = [[1, 0, 0]]
        >>> sd2 = [[0, 1, 0]]
        >>> sd3 = [[0, 0, 1]]
        >>> # Default units are meV
        >>> energy(sd1), energy(sd2), energy(sd3)
        (-4.5, -4.5, -6.75)
        >>> # You can request other units
        >>> print(f"{energy(sd1, units='Joule'):.4e}")
        -7.2098e-22
    """

    def __init__(self, spinham):
        initial_units = spinham.units
        initial_convention = spinham.convention

        magnopy_convention = initial_convention.get_modified(
            spin_normalized=False, multiple_counting=True
        )

        spinham.units = "meV"
        spinham.convention = magnopy_convention

        self.spins = np.array(spinham.magnetic_atoms.spins, dtype=float)
        self.M = spinham.M

        ########################################################################
        #                               One spin                               #
        ########################################################################

        self.J_1 = np.zeros((spinham.M, 3), dtype=float)

        for atom, parameter in spinham.p1:
            alpha = spinham.map_to_magnetic[atom]

            self.J_1[alpha] += spinham.convention.c1 * parameter

        ########################################################################
        #                               Two spins                              #
        ########################################################################

        self.J_21 = np.zeros((spinham.M, 3, 3), dtype=float)

        for atom, parameter in spinham.p21:
            alpha = spinham.map_to_magnetic[atom]

            self.J_21[alpha] += spinham.convention.c21 * parameter

        self.J_22 = {}

        for atom1, atom2, _, parameter in spinham.p22:
            alpha = spinham.map_to_magnetic[atom1]
            beta = spinham.map_to_magnetic[atom2]

            if (alpha, beta) not in self.J_22:
                self.J_22[(alpha, beta)] = np.zeros((3, 3), dtype=float)

            self.J_22[(alpha, beta)] += spinham.convention.c22 * parameter

        ########################################################################
        #                              Three spins                             #
        ########################################################################

        self.J_31 = np.zeros((spinham.M, 3, 3, 3), dtype=float)

        for atom, parameter in spinham.p31:
            alpha = spinham.map_to_magnetic[atom]

            self.J_31[alpha] += spinham.convention.c31 * parameter

        self.J_32 = {}

        for atom1, atom2, _, parameter in spinham.p32:
            alpha = spinham.map_to_magnetic[atom1]
            beta = spinham.map_to_magnetic[atom2]

            if (alpha, beta) not in self.J_32:
                self.J_32[(alpha, beta)] = np.zeros((3, 3, 3), dtype=float)

            self.J_32[(alpha, beta)] += spinham.convention.c32 * parameter

        self.J_33 = {}

        for atom1, atom2, atom3, _, _, parameter in spinham.p33:
            alpha = spinham.map_to_magnetic[atom1]
            beta = spinham.map_to_magnetic[atom2]
            gamma = spinham.map_to_magnetic[atom3]

            if (alpha, beta, gamma) not in self.J_33:
                self.J_33[(alpha, beta, gamma)] = np.zeros((3, 3, 3), dtype=float)

            self.J_33[(alpha, beta, gamma)] += spinham.convention.c33 * parameter

        ########################################################################
        #                              Four spins                              #
        ########################################################################

        self.J_41 = np.zeros((spinham.M, 3, 3, 3, 3), dtype=float)

        for atom, parameter in spinham.p41:
            alpha = spinham.map_to_magnetic[atom]

            self.J_41[alpha] += spinham.convention.c41 * parameter

        self.J_421 = {}

        for atom1, atom2, _, parameter in spinham.p421:
            alpha = spinham.map_to_magnetic[atom1]
            beta = spinham.map_to_magnetic[atom2]

            if (alpha, beta) not in self.J_421:
                self.J_421[(alpha, beta)] = np.zeros((3, 3, 3, 3), dtype=float)

            self.J_421[(alpha, beta)] += spinham.convention.c421 * parameter

        self.J_422 = {}

        for atom1, atom2, _, parameter in spinham.p422:
            alpha = spinham.map_to_magnetic[atom1]
            beta = spinham.map_to_magnetic[atom2]

            if (alpha, beta) not in self.J_422:
                self.J_422[(alpha, beta)] = np.zeros((3, 3, 3, 3), dtype=float)

            self.J_422[(alpha, beta)] += spinham.convention.c422 * parameter

        self.J_43 = {}

        for atom1, atom2, atom3, _, _, parameter in spinham.p43:
            alpha = spinham.map_to_magnetic[atom1]
            beta = spinham.map_to_magnetic[atom2]
            gamma = spinham.map_to_magnetic[atom3]

            if (alpha, beta, gamma) not in self.J_43:
                self.J_43[(alpha, beta, gamma)] = np.zeros((3, 3, 3, 3), dtype=float)

            self.J_43[(alpha, beta, gamma)] += spinham.convention.c43 * parameter

        self.J_44 = {}

        for atom1, atom2, atom3, atom4, _, _, _, parameter in spinham.p44:
            alpha = spinham.map_to_magnetic[atom1]
            beta = spinham.map_to_magnetic[atom2]
            gamma = spinham.map_to_magnetic[atom3]
            epsilon = spinham.map_to_magnetic[atom4]

            if (alpha, beta, gamma, epsilon) not in self.J_44:
                self.J_44[(alpha, beta, gamma, epsilon)] = np.zeros(
                    (3, 3, 3, 3), dtype=float
                )

            self.J_44[(alpha, beta, gamma, epsilon)] += (
                spinham.convention.c44 * parameter
            )

        spinham.units = initial_units
        spinham.convention = initial_convention

    def __call__(self, spin_directions, units="meV", _normalize=True) -> float:
        return self.E_0(
            spin_directions=spin_directions, units=units, _normalize=_normalize
        )

    def E_0(self, spin_directions, units="meV", _normalize=True) -> float:
        r"""
        Computes classical energy of the spin Hamiltonian.

        Parameters
        ----------

        spin_directions : (M, 3) |array-like|_
            Directions of spin vectors. Only directions of vectors are used, modulus is
            ignored. ``M`` is the amount of magnetic atoms in the Hamiltonian. The order
            of spin directions is the same as the order of magnetic atoms in
            ``spinham.magnetic_atoms.spins``.

        units : str, default "meV"
            .. versionadded:: 0.3.0

            Units of energy. See :ref:`user-guide_usage_units_energy-units` for the full
            list of supported units.


        _normalize : bool, default True
            Whether to normalize the spin_directions or use the provided vectors as is.
            This parameter is technical and we do not recommend to use it at all.

        Returns
        -------

        E_0 : float
            Classic energy of state with ``spin_directions``. Returned in the units of
            ``units``.


        Examples
        --------

        First, one need to create some spin Hamiltonian

        .. doctest::

            >>> import numpy as np
            >>> import magnopy
            >>> cell = np.eye(3)
            >>> atoms = dict(
            ...     names=["Fe"], spins=[1.5], g_factors=[2], positions=[[0, 0, 0]]
            ... )
            >>> convention = magnopy.Convention(
            ...     multiple_counting=True, spin_normalized=False, c21=1, c22=-1
            ... )
            >>> spinham = magnopy.SpinHamiltonian(
            ...     cell=cell, atoms=atoms, convention=convention
            ... )

        Then, add some parameters to the Hamiltonian

        .. doctest::

            >>> spinham.add_21(alpha=0, parameter=np.diag([0, 0, -1]))
            >>> spinham.add_22(
            ...     alpha=0,
            ...     beta=0,
            ...     nu=(1, 0, 0),
            ...     parameter=magnopy.converter22.from_iso(iso=1),
            ... )

        Now everything is ready to create an instance of the Energy class

        .. doctest::

            >>> energy = magnopy.Energy(spinham)

        Finally, ``energy`` an be used to compute classical energy of the Hamiltonian
        for arbitrary spin configuration.

        .. doctest::

            >>> sd1 = [[1, 0, 0]]
            >>> sd2 = [[0, 1, 0]]
            >>> sd3 = [[0, 0, 1]]
            >>> # Default units are meV
            >>> energy.E_0(sd1), energy.E_0(sd2), energy.E_0(sd3)
            (-4.5, -4.5, -6.75)
            >>> # The command above is equivalent to
            >>> energy(sd1), energy(sd2), energy(sd3)
            (-4.5, -4.5, -6.75)
        """

        spin_directions = np.array(spin_directions, dtype=float)

        if _normalize:
            spin_directions = (
                spin_directions / np.linalg.norm(spin_directions, axis=1)[:, np.newaxis]
            )
        spins = spin_directions * self.spins[:, np.newaxis]

        energy = 0

        energy += np.diag(self.J_1 @ spins.T).sum()

        energy += np.einsum("mij,mi,mj->m", self.J_21, spins, spins).sum()

        energy += np.einsum("miju,mi,mj,mu->m", self.J_31, spins, spins, spins).sum()

        energy += np.einsum(
            "mijuv,mi,mj,mu,mv->m", self.J_41, spins, spins, spins, spins
        ).sum()

        for alpha, beta in self.J_22:
            energy += spins[alpha] @ self.J_22[(alpha, beta)] @ spins[beta]

        for alpha, beta in self.J_32:
            energy += np.einsum(
                "iju,i,j,u",
                self.J_32[(alpha, beta)],
                spins[alpha],
                spins[alpha],
                spins[beta],
            )

        for alpha, beta in self.J_421:
            energy += np.einsum(
                "ijuv,i,j,u,v",
                self.J_421[(alpha, beta)],
                spins[alpha],
                spins[alpha],
                spins[alpha],
                spins[beta],
            )

        for alpha, beta in self.J_422:
            energy += np.einsum(
                "ijuv,i,j,u,v",
                self.J_422[(alpha, beta)],
                spins[alpha],
                spins[alpha],
                spins[beta],
                spins[beta],
            )

        for alpha, beta, gamma in self.J_33:
            energy += np.einsum(
                "iju,i,j,u",
                self.J_33[(alpha, beta, gamma)],
                spins[alpha],
                spins[beta],
                spins[gamma],
            )

        for alpha, beta, gamma in self.J_43:
            energy += np.einsum(
                "ijuv,i,j,u,v",
                self.J_43[(alpha, beta, gamma)],
                spins[alpha],
                spins[alpha],
                spins[beta],
                spins[gamma],
            )

        for alpha, beta, gamma, epsilon in self.J_44:
            energy += np.einsum(
                "ijuv,i,j,u,v",
                self.J_44[(alpha, beta, gamma, epsilon)],
                spins[alpha],
                spins[beta],
                spins[gamma],
                spins[epsilon],
            )

        # Convert units if necessary
        if units != "meV":
            units = _validated_units(units=units, supported_units=_ENERGY_UNITS)
            energy = energy * _ENERGY_UNITS["mev"] / _ENERGY_UNITS[units]

        return float(energy)

    def gradient(self, spin_directions, units="meV", _normalize=True):
        r"""
        Computes first derivatives of energy (:math:`E^{(0)}`) with respect to the
        components of the spin directional vectors.

        Parameters
        ----------

        spin_directions : (M, 3) |array-like|_
            Directions of spin vectors. Only directions of vectors are used,
            modulus is ignored. ``M`` is the amount of magnetic atoms in the
            Hamiltonian. The order of spin directions is the same as the order
            of magnetic atoms in ``spinham.magnetic_atoms.spins``.

        units : str, default "meV"
            .. versionadded:: 0.3.0

            Units of energy. See :ref:`user-guide_usage_units_energy-units` for the full
            list of supported units.


        _normalize : bool, default True
            Whether to normalize the spin_directions or use the provided vectors as is.
            This parameter is technical and we do not recommend to use it at all.

        Returns
        -------

        gradient : (M, 3) :numpy:`ndarray`
            Gradient of energy.

            .. code-block:: python

                [
                    [dE / dz1x, dE / dz1y, dE / dz1z],
                    [dE / dz2x, dE / dz2y, dE / dz2z],
                    ...[dE / dzMx, dE / dzMy, dE / dzMz],
                ]
        """

        spin_directions = np.array(spin_directions, dtype=float)

        if _normalize:
            spin_directions = (
                spin_directions / np.linalg.norm(spin_directions, axis=1)[:, np.newaxis]
            )

        gradient = np.zeros((self.M, 3), dtype=float)

        gradient += self.J_1 * self.spins[:, np.newaxis]

        gradient += 2 * np.einsum(
            "mtj,mj,m->mt", self.J_21, spin_directions, self.spins**2
        )

        gradient += 3 * np.einsum(
            "mtju,mj,mu,m->mt",
            self.J_31,
            spin_directions,
            spin_directions,
            self.spins**3,
        )

        gradient += 4 * np.einsum(
            "mtjuv,mj,mu,mv,m->mt",
            self.J_41,
            spin_directions,
            spin_directions,
            spin_directions,
            self.spins**4,
        )

        for alpha, beta in self.J_22:
            gradient[alpha] += 2 * (
                self.J_22[(alpha, beta)]
                @ spin_directions[beta]
                * self.spins[alpha]
                * self.spins[beta]
            )

        for alpha, beta in self.J_32:
            gradient[alpha] += 3 * (
                np.einsum(
                    "tju,j,u->t",
                    self.J_32[(alpha, beta)],
                    spin_directions[alpha],
                    spin_directions[beta],
                )
                * self.spins[alpha] ** 2
                * self.spins[beta]
            )

        for alpha, beta in self.J_421:
            gradient[alpha] += 4 * (
                np.einsum(
                    "tjuv,j,u,v->t",
                    self.J_421[(alpha, beta)],
                    spin_directions[alpha],
                    spin_directions[alpha],
                    spin_directions[beta],
                )
                * self.spins[alpha] ** 3
                * self.spins[beta]
            )

        for alpha, beta in self.J_422:
            gradient[alpha] += 4 * (
                np.einsum(
                    "tjuv,j,u,v->t",
                    self.J_422[(alpha, beta)],
                    spin_directions[alpha],
                    spin_directions[beta],
                    spin_directions[beta],
                )
                * self.spins[alpha] ** 2
                * self.spins[beta] ** 2
            )

        for alpha, beta, gamma in self.J_33:
            gradient[alpha] += 3 * (
                np.einsum(
                    "tju,j,u->t",
                    self.J_33[(alpha, beta, gamma)],
                    spin_directions[beta],
                    spin_directions[gamma],
                )
                * self.spins[alpha]
                * self.spins[beta]
                * self.spins[gamma]
            )

        for alpha, beta, gamma in self.J_43:
            gradient[alpha] += 4 * (
                np.einsum(
                    "tjuv,j,u,v->t",
                    self.J_43[(alpha, beta, gamma)],
                    spin_directions[alpha],
                    spin_directions[beta],
                    spin_directions[gamma],
                )
                * self.spins[alpha] ** 2
                * self.spins[beta]
                * self.spins[gamma]
            )

        for alpha, beta, gamma, epsilon in self.J_44:
            gradient[alpha] += 4 * (
                np.einsum(
                    "tjuv,j,u,v->t",
                    self.J_44[(alpha, beta, gamma, epsilon)],
                    spin_directions[beta],
                    spin_directions[gamma],
                    spin_directions[epsilon],
                )
                * self.spins[alpha]
                * self.spins[beta]
                * self.spins[gamma]
                * self.spins[epsilon]
            )

        # Convert units if necessary
        if units != "meV":
            units = _validated_units(units=units, supported_units=_ENERGY_UNITS)
            gradient = gradient * _ENERGY_UNITS["mev"] / _ENERGY_UNITS[units]

        return gradient

    def torque(self, spin_directions, units="meV", _normalize=True):
        r"""
        Computes torque on each spin.

        Parameters
        ----------

        spin_directions : (M, 3) |array-like|_
            Directions of spin vectors. Only directions of vectors are used,
            modulus is ignored. ``M`` is the amount of magnetic atoms in the
            Hamiltonian. The order of spin directions is the same as the order
            of magnetic atoms in ``spinham.magnetic_atoms.spins``.

        units : str, default "meV"
            .. versionadded:: 0.3.0

            Units of energy. See :ref:`user-guide_usage_units_energy-units` for the full
            list of supported units.

        _normalize : bool, default True
            Whether to normalize the spin_directions or use the provided vectors as is.
            This parameter is technical and we do not recommend to use it at all.

        Returns
        -------

        torque : (M, 3) :numpy:`ndarray`

            .. code-block:: python

                [[t1x, t1y, t1z], [t2x, t2y, t2z], ...[tMx, tMy, tMz]]
        """
        return np.cross(
            spin_directions,
            self.gradient(
                spin_directions=spin_directions, units=units, _normalize=_normalize
            ),
        )

    def _zoom(
        self,
        reference_sd,
        search_direction,
        phi_0,
        der_0,
        alpha_lo,
        alpha_hi,
        c1=_C1,
        c2=_C2,
    ):
        sd_lo = _rotate_sd(
            reference_sd=reference_sd, rotation=alpha_lo * search_direction
        )
        sd_hi = _rotate_sd(
            reference_sd=reference_sd, rotation=alpha_hi * search_direction
        )

        phi_lo = self.E_0(spin_directions=sd_lo)
        phi_hi = self.E_0(spin_directions=sd_hi)

        der_lo = self.torque(spin_directions=sd_lo).flatten() @ search_direction
        der_hi = self.torque(spin_directions=sd_hi).flatten() @ search_direction

        trial_steps = 0
        phi_min = None
        while True:
            # Interpolate
            alpha_j = _cubic_interpolation(
                alpha_l=alpha_lo,
                alpha_h=alpha_hi,
                phi_l=phi_lo,
                phi_h=phi_hi,
                der_l=der_lo,
                der_h=der_hi,
            )
            # Evaluate \phi(\alpha_i)
            sd_j = _rotate_sd(
                reference_sd=reference_sd, rotation=alpha_j * search_direction
            )
            phi_j = self.E_0(spin_directions=sd_j)

            # Safeguard
            if phi_min is None:
                phi_min = phi_j
            else:
                if phi_j < phi_min:
                    phi_min = phi_j
                    trial_steps = 0
            trial_steps += 1
            if trial_steps > 10:
                return alpha_j

            # Evaluate \phi^{\prime}(\alpha_i)
            der_j = self.torque(spin_directions=sd_j).flatten() @ search_direction

            if phi_j > phi_0 + c1 * alpha_j * der_0 or phi_j >= phi_lo:
                alpha_hi = alpha_j
                phi_hi = phi_j
                der_hi = der_j
            else:
                if abs(der_j) <= -c2 * der_0:
                    return alpha_j
                if der_j * (alpha_hi - alpha_lo) >= 0:
                    alpha_hi = alpha_lo
                    phi_hi = phi_lo
                    der_hi = der_lo
                alpha_lo = alpha_j
                phi_lo = phi_j
                der_lo = der_j

    def _line_search(
        self,
        reference_sd,
        search_direction,
        phi_0,
        der_0,
        c1=_C1,
        c2=_C2,
        alpha_max=2.0,
        max_iterations=10000,
    ):
        # First check if step alpha=1 is good to go:
        sd_1 = _rotate_sd(reference_sd=reference_sd, rotation=search_direction)
        phi_1 = self.E_0(spin_directions=sd_1)
        der_1 = self.torque(spin_directions=sd_1).flatten() @ search_direction

        if phi_1 <= phi_0 + c1 * der_0 and abs(der_1) <= c2 * abs(der_0):
            return 1.0

        # If not, then proceed to the algorithm

        alpha_prev = 0
        phi_prev = phi_0
        der_prev = der_0

        sd_max = _rotate_sd(
            reference_sd=reference_sd, rotation=alpha_max * search_direction
        )
        phi_max = self.E_0(spin_directions=sd_max)
        der_max = self.torque(spin_directions=sd_max).flatten() @ search_direction

        alpha_i = _cubic_interpolation(
            alpha_l=alpha_prev,
            alpha_h=alpha_max,
            phi_l=phi_prev,
            phi_h=phi_max,
            der_l=der_prev,
            der_h=der_max,
        )

        for i in range(1, max_iterations):
            # Evaluate \phi(\alpha_i)
            sd_i = _rotate_sd(
                reference_sd=reference_sd, rotation=alpha_i * search_direction
            )
            phi_i = self.E_0(spin_directions=sd_i)

            if phi_i > phi_0 + c1 * alpha_i * der_0 or (i > 1 and phi_i >= phi_prev):
                return self._zoom(
                    reference_sd=reference_sd,
                    search_direction=search_direction,
                    phi_0=phi_0,
                    der_0=der_0,
                    alpha_lo=alpha_prev,
                    alpha_hi=alpha_i,
                )

            # Evaluate \phi^{\prime}(\alpha_i)
            der_i = self.torque(spin_directions=sd_i).flatten() @ search_direction

            if abs(der_i) <= -c2 * der_0:
                return alpha_i

            if der_i >= 0:
                return self._zoom(
                    reference_sd=reference_sd,
                    search_direction=search_direction,
                    phi_0=phi_0,
                    der_0=der_0,
                    alpha_lo=alpha_i,
                    alpha_hi=alpha_prev,
                )

            # Choose alpha_{i+1}
            alpha_next = _cubic_interpolation(
                alpha_l=alpha_i,
                alpha_h=alpha_max,
                phi_l=phi_i,
                phi_h=phi_max,
                der_l=der_i,
                der_h=der_max,
            )

            # Set i <- i+1
            alpha_prev = alpha_i
            phi_prev = phi_i
            der_prev = der_i

            alpha_i = alpha_next

        raise ValueError(
            f"Line search did not converge in {max_iterations} iterations."
        )

    def optimize(
        self,
        initial_guess=None,
        energy_tolerance=1e-5,
        torque_tolerance=1e-5,
        quiet=False,
    ):
        r"""
        Optimizes classical energy by varying the directions of spins in the unit cell.

        Parameters
        ----------

        initial_guess : (M, 3) or (3,) |array-like|_, optional
            Initial guess for the direction of the spin vectors.

        energy_tolerance : float, default 1e-5
            Energy tolerance for the two consecutive steps of the optimization. In the units of meV.

        torque_tolerance : float, default 1e-5
            Torque tolerance for the two consecutive steps of the optimization. In the units of meV.

        quiet : bool, default False
            Whether to suppress the output of the progress.

        Returns
        -------

        optimized_directions : (M, 3) :numpy:`ndarray`
            Optimized direction of the spin vectors.

        See Also
        --------

        optimize_generator
        """

        if initial_guess is None:
            initial_guess = np.random.uniform(low=-1, high=1, size=(self.M, 3))

        sd_k = initial_guess / np.linalg.norm(initial_guess, axis=1)[:, np.newaxis]

        if not quiet:
            n_energy = max(-(int(log10(energy_tolerance)) - 2), 6)
            n_torque = max(-(int(log10(torque_tolerance)) - 2), 6)
            print(
                "─" * 5
                + "┬"
                + "─" * 13
                + "┬"
                + "─" * (6 + n_energy)
                + "┬"
                + "─" * (6 + n_torque)
            )
            print(
                f"{'step':^4} │ "
                f"{'E_0':^11} │ "
                f"{'delta E_0':^{n_energy + 4}} │ "
                f"{'max torque':^{n_torque + 4}}"
            )
            print(
                "─" * 5
                + "┴"
                + "─" * 13
                + "┴"
                + "─" * (6 + n_energy)
                + "┴"
                + "─" * (6 + n_torque)
            )

        tolerance = np.array([energy_tolerance, torque_tolerance], dtype=float)

        delta = 2 * tolerance

        hessinv_k = np.eye(3 * self.M, dtype=float)

        energy_k = self.E_0(spin_directions=sd_k)
        gradient_k = self.torque(spin_directions=sd_k).flatten()

        first_iteration = True
        step_counter = 1
        # Curvature failure guard to avoid silent looping in degenerate manifolds
        curv_fail_run = 0
        max_curv_fails = 10

        while (delta >= tolerance).any():
            search_direction = -hessinv_k @ gradient_k

            alpha_k = self._line_search(
                reference_sd=sd_k,
                search_direction=search_direction,
                phi_0=energy_k,
                der_0=gradient_k @ search_direction,
            )

            s_k = alpha_k * search_direction

            sd_next = _rotate_sd(reference_sd=sd_k, rotation=s_k)

            energy_next = self.E_0(spin_directions=sd_next)
            gradient_next = self.torque(spin_directions=sd_next).flatten()

            delta = np.array(
                [
                    abs(energy_next - energy_k),
                    # Pay attention to the np.reshape keywords
                    np.linalg.norm(
                        np.reshape(gradient_next, (self.M, 3)), axis=1
                    ).max(),
                ]
            )

            if not quiet:
                print(
                    f"{step_counter:<4}   "
                    f"{energy_next:>11.7f}   "
                    f"{delta[0]:>{n_energy + 4}.{n_energy}f}   "
                    f"{delta[1]:>{n_torque + 4}.{n_torque}f}"
                )

            if (delta < tolerance).all():
                break

            y_k = gradient_next - gradient_k
            # Curvature safeguard: avoid divide-by-zero / non-finite ys
            ys = float(y_k @ s_k)
            EYE = np.eye(hessinv_k.shape[0], dtype=float)

            if (not np.isfinite(ys)) or (abs(ys) < 1e-12):
                # Degenerate case: skip the inverse-BFGS update (or reset H if preferred)
                curv_fail_run += 1
                if curv_fail_run > max_curv_fails:
                    raise RuntimeError(
                        f"BFGS curvature failure repeated {curv_fail_run} times: "
                        f"s^T y={ys}, ||s||={np.linalg.norm(s_k)}, ||y||={np.linalg.norm(y_k)}. "
                        "Try a different initial guess (i.e. re-run magnopy-optimize-sd or energy.optimize())."
                    )
                # Reset hessian:
                hessinv_k = EYE
                warnings.warn(
                    f"BFGS curvature failure repeated {curv_fail_run} times: s^T y={ys}, ||s||={np.linalg.norm(s_k)}, ||y||={np.linalg.norm(y_k)}. Hessian was reset.",
                    RuntimeWarning,
                    stacklevel=2,
                )
            else:
                curv_fail_run = 0
                rho_k = 1.0 / ys
                OUTER = np.outer(y_k, s_k)

                # Safe initial scaling of H^{-1}
                if first_iteration:
                    first_iteration = False
                    denom = float(y_k @ y_k)
                    if np.isfinite(denom) and denom > 0.0:
                        hessinv_k = (ys / denom) * hessinv_k

                # Stable inverse-BFGS update
                hessinv_k = (EYE - rho_k * OUTER.T) @ hessinv_k @ (
                    EYE - rho_k * OUTER
                ) + rho_k * np.outer(s_k, s_k)

            sd_k = sd_next
            energy_k = energy_next
            gradient_k = gradient_next

            step_counter += 1
            # print("=" * 40)

        if not quiet:
            print("─" * (33 + n_energy + n_torque))
        return sd_next

    def optimize_generator(
        self,
        initial_guess=None,
        energy_tolerance=1e-5,
        torque_tolerance=1e-5,
    ):
        r"""
        Optimizes classical energy by varying the directions of spins in the unit cell.

        .. versionadded:: 0.2.0

        .. warning::
            This method is experimental, use at your own risk. Use
            :py:meth:`.Energy.optimize` as a stable alternative.

        Parameters
        ----------

        initial_guess : (M, 3) or (3,) |array-like|_, optional
            Initial guess for the direction of the spin vectors.

        energy_tolerance : float, default 1e-5
            Energy tolerance for the two consecutive steps of the optimization.

        torque_tolerance : float, default 1e-5
            Torque tolerance for the two consecutive steps of the optimization.

        Yields
        ------

        energy : float
            Classical energy of the iteration step

        gradient : (M, 3) :numpy:`ndarray`
            Gradient vectors for each spin of the iteration step.

        spin_directions : (M, 3) :numpy:`ndarray`
            Directions of the spin vectors of the iteration step.

        See Also
        --------

        optimize
        """

        if initial_guess is None:
            initial_guess = np.random.uniform(low=-1, high=1, size=(self.M, 3))

        sd_k = initial_guess / np.linalg.norm(initial_guess, axis=1)[:, np.newaxis]

        tolerance = np.array([energy_tolerance, torque_tolerance], dtype=float)

        delta = 2 * tolerance

        hessinv_k = np.eye(3 * self.M, dtype=float)

        energy_k = self.E_0(spin_directions=sd_k)
        gradient_k = self.torque(spin_directions=sd_k).flatten()

        first_iteration = True
        step_counter = 1
        # Curvature failure guard to avoid silent looping in degenerate manifolds
        curv_fail_run = 0
        max_curv_fails = 10

        yield (energy_k, gradient_k, sd_k)

        while (delta >= tolerance).any():
            search_direction = -hessinv_k @ gradient_k

            alpha_k = self._line_search(
                reference_sd=sd_k,
                search_direction=search_direction,
                phi_0=energy_k,
                der_0=gradient_k @ search_direction,
            )

            s_k = alpha_k * search_direction

            sd_next = _rotate_sd(reference_sd=sd_k, rotation=s_k)

            energy_next = self.E_0(spin_directions=sd_next)
            gradient_next = self.torque(spin_directions=sd_next).flatten()

            yield (energy_next, gradient_next, sd_next)

            delta = np.array(
                [
                    abs(energy_next - energy_k),
                    # Pay attention to the np.reshape keywords
                    np.linalg.norm(
                        np.reshape(gradient_next, (self.M, 3)), axis=1
                    ).max(),
                ]
            )

            if (delta < tolerance).all():
                break

            y_k = gradient_next - gradient_k
            # Curvature safeguard: avoid divide-by-zero / non-finite ys
            ys = float(y_k @ s_k)
            EYE = np.eye(hessinv_k.shape[0], dtype=float)

            if (not np.isfinite(ys)) or (abs(ys) < 1e-12):
                # Degenerate case: skip the inverse-BFGS update (or reset H if preferred)
                curv_fail_run += 1
                if curv_fail_run > max_curv_fails:
                    raise RuntimeError(
                        f"BFGS curvature failure repeated {curv_fail_run} times: "
                        f"s^T y={ys}, ||s||={np.linalg.norm(s_k)}, ||y||={np.linalg.norm(y_k)}. "
                        "Try a different initial guess (i.e. re-run magnopy-optimize-sd or energy.optimize())."
                    )
                # Reset hessian:
                hessinv_k = EYE
                warnings.warn(
                    f"BFGS curvature failure repeated {curv_fail_run} times: s^T y={ys}, ||s||={np.linalg.norm(s_k)}, ||y||={np.linalg.norm(y_k)}. Hessian was reset.",
                    RuntimeWarning,
                    stacklevel=2,
                )
            else:
                curv_fail_run = 0
                rho_k = 1.0 / ys
                OUTER = np.outer(y_k, s_k)

                # Safe initial scaling of H^{-1}
                if first_iteration:
                    first_iteration = False
                    denom = float(y_k @ y_k)
                    if np.isfinite(denom) and denom > 0.0:
                        hessinv_k = (ys / denom) * hessinv_k

                # Stable inverse-BFGS update
                hessinv_k = (EYE - rho_k * OUTER.T) @ hessinv_k @ (
                    EYE - rho_k * OUTER
                ) + rho_k * np.outer(s_k, s_k)

            sd_k = sd_next
            energy_k = energy_next
            gradient_k = gradient_next

            step_counter += 1

        return sd_next


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _envelope_warning(text):
    lines = text.split("\n")

    short_lines = []
    for line in lines:
        if len(line) > 74:
            line = line.split()
            while len(line) > 0:
                subline = line[0]
                line = line[1:]
                while len(subline) < 74 and len(line) > 0:
                    if len(subline) + 1 + len(line[0]) <= 74:
                        subline = f"{subline} {line[0]}"
                        line = line[1:]
                    else:
                        break
                short_lines.append(subline)
        else:
            short_lines.append(line)

    enveloped_lines = [f"#W {line:<74} W#" for line in short_lines]

    return (
        f"\n{'  WARNING  ':#^80}\n"
        + "\n".join(enveloped_lines)
        + f"\n{'  END OF WARNING  ':#^80}\n"
    )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
from magnopy._constants._si import (
    BOLTZMANN_CONSTANT,
    ELECTRON_VOLT,
    JOULE,
    KELVIN,
    MILLI,
    RYDBERG_ENERGY,
    ERG,
    GIGA,
    TERA,
    PLANCK_CONSTANT,
)


################################################################################
#                                    Energy                                    #
################################################################################

# Name : value when expressed in SI
_ENERGY_UNITS = {
    "ev": ELECTRON_VOLT,
    "mev": MILLI * ELECTRON_VOLT,
    "joule": JOULE,
    "j": JOULE,
    "ry": RYDBERG_ENERGY,
    "rydberg": RYDBERG_ENERGY,
    "erg": ERG,
}

# Name : Pretty name
_ENERGY_UNITS_MAKEUP = {
    "ev": "eV",
    "mev": "meV",
    "joule": "Joule",
    "j": "Joule",
    "ry": "Rydberg",
    "rydberg": "Rydberg",
    "erg": "Erg",
}

################################################################################
#                         Parameter of spin Hamiltonian                        #
################################################################################

# Name : value when expressed in SI
_PARAMETER_UNITS = {key: _ENERGY_UNITS[key] for key in _ENERGY_UNITS}
_PARAMETER_UNITS["k"] = BOLTZMANN_CONSTANT * KELVIN
_PARAMETER_UNITS["kelvin"] = BOLTZMANN_CONSTANT * KELVIN


# Name : Pretty name
_PARAMETER_UNITS_MAKEUP = {
    key: _ENERGY_UNITS_MAKEUP[key] for key in _ENERGY_UNITS_MAKEUP
}
_PARAMETER_UNITS_MAKEUP["k"] = "Kelvin"
_PARAMETER_UNITS_MAKEUP["kelvin"] = "Kelvin"

################################################################################
#                              Magnon frequencies                              #
################################################################################

# Name : value when expressed in SI
_MAGNON_ENERGY_UNITS = {key: _ENERGY_UNITS[key] for key in _ENERGY_UNITS}
_MAGNON_ENERGY_UNITS["hertz"] = PLANCK_CONSTANT
_MAGNON_ENERGY_UNITS["hz"] = PLANCK_CONSTANT
_MAGNON_ENERGY_UNITS["gigahertz"] = PLANCK_CONSTANT * GIGA
_MAGNON_ENERGY_UNITS["ghz"] = PLANCK_CONSTANT * GIGA
_MAGNON_ENERGY_UNITS["terahertz"] = PLANCK_CONSTANT * TERA
_MAGNON_ENERGY_UNITS["thz"] = PLANCK_CONSTANT * TERA


# Name : Pretty name
_MAGNON_ENERGY_UNITS_MAKEUP = {
    key: _ENERGY_UNITS_MAKEUP[key] for key in _ENERGY_UNITS_MAKEUP
}
_MAGNON_ENERGY_UNITS_MAKEUP["hertz"] = "Hz"
_MAGNON_ENERGY_UNITS_MAKEUP["hz"] = "Hz"
_MAGNON_ENERGY_UNITS_MAKEUP["gigahertz"] = "GHz"
_MAGNON_ENERGY_UNITS_MAKEUP["ghz"] = "GHz"
_MAGNON_ENERGY_UNITS_MAKEUP["terahertz"] = "THz"
_MAGNON_ENERGY_UNITS_MAKEUP["thz"] = "THz"



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


from typing import Iterable
import warnings
import numpy as np

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


# DEPRECATED in v0.2.0
# Remove in March 2026
def output_k_resolved(
    data,
    data_headers=None,
    output_filename=None,
    kpoints=None,
    relative=False,
    rcell=None,
    flat_indices=None,
    digits=4,
    scientific_notation=True,
):
    r"""
    Outputs any k-resolved data.

    .. deprecated:: 0.2.0
        Always meant to be a temporary solution. This function will be removed in March of 2026


    Parameters
    ----------
    data : (N, M) |array-like|_
        Some k-resolved data. N (:math:`\ge 1`) is the amount of kpoints. M is the
        number of data modes/entries.
    data_headers : (N, ) list of str, optional
        Header for the data columns. ``[f"data {i+1} for i in range(len(data))]`` by
        default.
    output_filename : str, optional
        Name of the file for saving the data. If ``None``, then return a list of lines.
    kpoints : (N, 3) |array-like|_, optional
        List of the kpoints. ``kpoints[i]`` correspond to the ``data[i]``.
    relative : bool, default False
        If ``relative == True``, then given ``kpoints`` are interpreted as relative to
        the reciprocal unit cell. Otherwise they are interpreted as absolute coordinates.
    rcell : (3, 3) |array-like|_, optional
        Reciprocal unit cell. Rows are interpreted as vectors, columns as Cartesian
        coordinates.
    flat_indices : (N, 3) |array-like|_, optional
        kpoints converted to the list of floats according to some rule. Typically used
        for plotting the band structure. If ``None`` provided, then computed automatically,
        based on the distance of the given ``kpoints`` (whether absolute or relative).
    digits : int, default 4
        Number of digits after the comma for the data elements.
    scientific_notation : bool, default True
        Whether to use scientific notation for the data elements.

    Returns
    -------
    lines : str
        Only returned if ``output_filename is None``. Use ``print("\n".join(lines))``
        to output the results to the standard output stream.
    """

    warnings.warn(
        "This function was deprecated in the release v0.2.0. This function will be removed from magnopy in March of 2026",
        DeprecationWarning,
        stacklevel=2,
    )

    # Prepare format for the data elements
    chars = digits + 1 + 1 + 3
    if scientific_notation:
        chars += 1 + 1 + 3
    chars = max(chars, 8)

    # Format string for data elements
    if scientific_notation:
        fmt = f">{chars}.{digits}e"
    else:
        fmt = f">{chars}.{digits}f"

    # Format string for kpoints
    kp_fmt = ">12.8f"

    # Prepare the header
    header = ["#"]
    if flat_indices is not None or kpoints is not None:
        header.append(f"{'flat index':>12}")
    header = header + [f"{tmp:>{chars}}" for tmp in data_headers]

    if kpoints is not None:
        if rcell is not None:
            header.append(
                " ".join([f"{f'k_{comp}':>12}" for comp in ["b1", "b2", "b3"]])
            )
            header.append(" ".join([f"{f'k_{comp}':>12}" for comp in "xyz"]))
        elif relative:
            header.append(
                " ".join([f"{f'k_{comp}':>12}" for comp in ["b1", "b2", "b3"]])
            )
        else:
            header.append(" ".join([f"{f'k_{comp}':>12}" for comp in "xyz"]))

    header = " ".join(header)

    # Output header
    if output_filename is not None:
        f = open(output_filename, "w", encoding="utf-8")
        f.write(header + "\n")
    else:
        lines = [header]

    # Output data for each kpoint
    for i in range(len(data)):
        line = [" "]

        if flat_indices is not None:
            line.append(f"{flat_indices[i]:{kp_fmt}}")
        elif kpoints is not None:
            if i == 0:
                line.append(f"{0.0:{kp_fmt}}")
            else:
                line.append(
                    f"{np.linalg.norm(np.array(kpoints[i], dtype=float) - kpoints[i - 1]):{kp_fmt}}"
                )

        if isinstance(data[i], Iterable):
            for entry in data[i]:
                line.append(f"{entry:{fmt}}")
        else:
            line.append(f"{data[i]:{fmt}}")

        if kpoints is not None:
            if rcell is not None:
                if relative:
                    k_vec = kpoints[i] @ np.array(rcell)
                    line.append(
                        " ".join([f"{kpoints[i][comp]:{kp_fmt}}" for comp in range(3)])
                    )
                    line.append(
                        " ".join([f"{k_vec[comp]:{kp_fmt}}" for comp in range(3)])
                    )
                else:
                    k_vec_rel = kpoints[i] @ np.linalg.inv(rcell)
                    line.append(
                        " ".join([f"{k_vec_rel[comp]:{kp_fmt}}" for comp in range(3)])
                    )
                    line.append(
                        " ".join([f"{kpoints[i][comp]:{kp_fmt}}" for comp in range(3)])
                    )
            else:
                line.append(
                    " ".join([f"{kpoints[i][comp]:{kp_fmt}}" for comp in range(3)])
                )

        line = " ".join(line)
        if output_filename is not None:
            f.write(line + "\n")
        else:
            lines.append(line)

    if output_filename is not None:
        f.close()
    else:
        return lines


# DEPRECATED in v0.3.0
# Remove in April 2026
def plot_k_resolved(data, kp=None, output_filename=None, ylabel=None):
    r"""
    Plots k-resolved data.

    .. deprecated:: 0.2.0
        Use :py:func:`magnopy.experimental.plot_dispersion` instead.
        :py:func:`magnopy.io.plot_k_resolved` will be removed in April of 2026.


    If only the ``data`` are given, then an index of the omegas is used for abscissa (x
    axis).

    Parameters
    ----------
    data : (N, M) |array-like|_
        Some k-resolved data. N (:math:`\ge 1`) is the amount of kpoints. M is the
        number of data modes/entries. Expected to be given in meV.
    kp : :py:class:`wulfric.Kpoints`, optional.
        Instance of the :py:class:`wulfric.Kpoints` class. It should be the same
        instance that were used in the preparation of the ``data``.
    output_filename : str, optional
        Name of the file for saving the image. If ``None``, then the graph would be
        opened in the interactive matplotlib window.
    ylabel : str, optional
        Label for the ordinate (y axis). Do not include units, units are included automatically.
    """
    import warnings

    warnings.warn(
        "Function magnopy.io.plot_k_resolved() was deprecated in 0.3.0 and will be removed in April 2026. Use magnopy.experimental.plot_dispersion() instead",
        DeprecationWarning,
        stacklevel=2,
    )

    from magnopy.experimental import plot_dispersion

    plot_dispersion(data=data, kp=kp, output_filename=output_filename, ylabel=ylabel)


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
from wulfric import PlotlyEngine as W_PlotlyEngine
import numpy as np
from random import choices
from string import ascii_lowercase as ASCII_LOWERCASE

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


class PlotlyEngine(W_PlotlyEngine):
    r"""
    Plotting engine based on |plotly|_.

    Parameters
    ----------

    fig : plotly graph object, optional
        Figure to plot on. If not provided, then a new one is created as
        ``fig = go.Figure()``.

    _sphinx_gallery_fix : bool, default  False
        Fixes display issues when building documentation using sphinx gallery.
        Please, always ignore this argument

    Attributes
    ----------

    fig : plotly graph object
        Figure to plot on.

    Notes
    -----

    This class is a part of ``magnopy[visual]`` and a child class of
    :py:class:`wulfric.PlotlyEngine`.
    """

    def plot_spin_directions(
        self,
        positions,
        spin_directions,
        colors="#000000",
        legend_label=None,
        legend_group=None,
        row=1,
        col=1,
    ):
        r"""
        Plots a set of spin directions.

        Parameters
        ----------

        positions : (N, 3) |array-like|_
            Positions of magnetic centers.

        spin_directions : (N, 3) |array-like|_
            Direction of spin vectors for each magnetic center. Only direction is used,
            the modulus is ignored. Spin directions are all normalized to the same length
            but not necessary to 1.

        colors : str or list of str, default "#000000"
            Color or colors for the arrows. Any value that is supported by |plotly|_.

        legend_label : str, optional
            Label of the line that is displayed in the figure.

        legend_group : str, optional
            Legend's group. If ``None``, then defaults to the random string of 10
            characters.

        row : int, default 1
            Row of the subplot.

        col : int, default 1
            Column of the subplot.
        """

        if legend_group is None:
            legend_group = "".join(choices(ASCII_LOWERCASE, k=10))

        pos = np.array(positions, dtype=float)
        sd = np.array(spin_directions, dtype=float)

        # Get normalization length
        tmp = pos[:, np.newaxis] - pos[np.newaxis, :]
        tmp = np.linalg.norm(tmp, axis=2)
        norm_length = (tmp + np.eye(tmp.shape[0]) * tmp.max()).min()

        if norm_length == 0:
            norm_length = 1

        if isinstance(colors, str):
            colors = [colors for _ in range(len(pos))]

        sd = sd / np.linalg.norm(sd, axis=1)[:, np.newaxis] * norm_length

        for i in range(len(pos)):
            self.plot_vector(
                start_point=pos[i],
                end_point=pos[i] + sd[i],
                color=colors[i],
                vector_label=None,
                legend_label=legend_label,
                legend_group=legend_group,
                row=row,
                col=col,
            )
            if legend_label is not None:
                legend_label = None


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


R"""
Units of the International system of units, physical constants expressed in SI, non-SI
units expressed in SI.
"""

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


################################################################################
##                              Seven base units                              ##
##                                                                            ##
##  Source The International System of Units (SI) 9th edition, Table 2        ##
##  Link: https://www.bipm.org/en/publications/si-brochure                    ##
##  Last updated: 22 September 2025                                           ##
################################################################################
SECOND = 1  # Time
METRE = 1  # Length
KILOGRAM = 1  # Mass
AMPERE = 1  # Electric current
KELVIN = 1  # Thermodynamic temperature
MOLE = 1  # amount of substance
CANDELA = 1  # Luminous intensity

################################################################################
##                           Twenty two named units                           ##
##                                                                            ##
##  Source The International System of Units (SI) 9th edition, Table 4        ##
##  Link: https://www.bipm.org/en/publications/si-brochure                    ##
##  Last updated: 22 September 2025                                           ##
################################################################################
RADIAN = 1  # Plane angle
STERADIAN = 1  # Solid angle
HERTZ = 1 / SECOND  # Frequency
NEWTON = KILOGRAM * METRE / SECOND**2  # Force
PASCAL = KILOGRAM / METRE / SECOND**2  # Pressure, stress
JOULE = KILOGRAM * METRE**2 / SECOND**2  # Energy, work, amount of heat
WATT = KILOGRAM * METRE**2 / SECOND**3  # Power, radiant flux
COULOMB = AMPERE * SECOND  # Electric charge
VOLT = KILOGRAM * METRE**2 / SECOND**3 / AMPERE  # Electric potential difference
FARAD = SECOND**4 * AMPERE**2 / KILOGRAM / METRE**2  # Capacitance
OHM = KILOGRAM * METRE**2 / SECOND**3 / AMPERE**2  # Electric resistance
SIEMENS = SECOND**3 * AMPERE**2 / KILOGRAM / METRE**2  # Electric conductance
WEBER = KILOGRAM * METRE**2 / SECOND**2 / AMPERE  # Magnetic flux
TESLA = KILOGRAM / SECOND**2 / AMPERE  # Magnetic flux density
HENRY = KILOGRAM * METRE**2 / SECOND**2 / AMPERE**2  # Inductance
DEGREE_CELSIUS = KELVIN  # Celsius temperature
LUMEN = CANDELA * STERADIAN  # Luminous flux
LUX = CANDELA * STERADIAN / METRE**2  # Illuminance
BECQUEREL = 1 / SECOND  # Activity referred to a radionuclide
GRAY = METRE**2 / SECOND**2  # Absorbed dose, kerma
SIEVERT = METRE**2 / SECOND**2  # Dose equivalent
KATAL = MOLE / SECOND  # Catalytic activity

################################################################################
##                                  Prefixes                                  ##
##                                                                            ##
##  Source The International System of Units (SI) 9th edition, Table 7        ##
##  Link: https://www.bipm.org/en/publications/si-brochure                    ##
##  Last updated: 22 September 2025                                           ##
################################################################################
DECA = 1e1
HECTO = 1e2
KILO = 1e3
MEGA = 1e6
GIGA = 1e9
TERA = 1e12
PETA = 1e15
EXA = 1e18
ZETTA = 1e21
YOTTA = 1e24
RONNA = 1e27
QUETTA = 1e30

DECI = 1e-1
CENTI = 1e-2
MILLI = 1e-3
MICRO = 1e-6
NANO = 1e-9
PICO = 1e-12
FEMTO = 1e-15
ATTO = 1e-18
ZEPTO = 1e-21
YOCTO = 1e-24
RONTO = 1e-27
QUECTO = 1e-30

################################################################################
##                          Seven defining constants                          ##
##                                                                            ##
##  Source The International System of Units (SI) 9th edition, Table 1        ##
##  Link: https://www.bipm.org/en/publications/si-brochure                    ##
##  Last updated: 22 September 2025                                           ##
################################################################################
HYPERFINE_TRANSITION_FREQUENCY_OF_CS = 9192631770 * HERTZ
SPEED_OF_LIGHT_IN_VACUUM = 299792458 * METRE / SECOND
PLANCK_CONSTANT = 6.62607015e-34 * JOULE * SECOND
ELEMENTARY_CHARGE = 1.602176634e-19 * COULOMB
BOLTZMANN_CONSTANT = 1.380649e-23 * JOULE / KELVIN
AVOGADRO_CONSTANT = 6.02214076e23 / MOLE
LUMINOUS_EFFICACY = 683 * LUMEN / WATT

################################################################################
##                             Physical constants                             ##
##                                                                            ##
##  Source: 2022 CODATA recommended values                                    ##
##  Link: https://pml.nist.gov/cuu/Constants/index.html                       ##
##  Last updated: 22 September 2025                                           ##
################################################################################
ELECTRON_VOLT = 1.602176634e-19 * JOULE
BOHR_RADIUS = 5.29177210544e-11 * METRE
BOHR_MAGNETON = 9.2740100657e-24 * JOULE / TESLA
RYDBERG_CONSTANT = 10973731.568157 / METRE
VACUUM_MAGNETIC_PERMEABILITY = 1.25663706127e-6 * NEWTON / AMPERE**2  # Newton/ Ampere^2
REDUCED_PLANCK_CONSTANT = (
    6.62607015e-34 / 2 / 3.14159265358979323846 * JOULE * SECOND / RADIAN
)

################################################################################
##                                Non-SI units                                ##
################################################################################
ANGSTROM = 1e-10 * METRE
RYDBERG_ENERGY = PLANCK_CONSTANT * SPEED_OF_LIGHT_IN_VACUUM * RYDBERG_CONSTANT * JOULE
ERG = 1e-7 * JOULE


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import os
import sys
from argparse import ArgumentParser, RawDescriptionHelpFormatter
import warnings

from magnopy._package_info import logo
from magnopy.io._grogu import load_grogu
from magnopy.io._tb2j import load_tb2j
from magnopy.scenarios._optimize_sd import optimize_sd
from magnopy._cli._arguments_library import (
    _get_command_info,
    _add_spinham_input,
    _add_spin_values,
    _add_magnetic_field,
    _add_output_folder,
    _add_no_html,
    _add_hide_personal_data,
    _add_make_sd_image,
    _add_supercell,
    _add_energy_tolerance,
    _add_torque_tolerance,
)
from magnopy._constants._icons import ICON_IN_FILE


def manager():
    # Configure a parser
    parser = ArgumentParser(
        description=logo()
        + "\n\nThis script optimizes classical energy of the spin Hamiltonian and "
        "finds spin directions that describe a local minima of the energy landscape.",
        formatter_class=RawDescriptionHelpFormatter,
    )

    _add_spinham_input(parser=parser)
    _add_energy_tolerance(parser=parser)
    _add_torque_tolerance(parser=parser)
    _add_supercell(parser=parser)
    _add_magnetic_field(parser=parser)
    _add_output_folder(parser=parser)
    _add_spin_values(parser=parser)
    _add_no_html(parser=parser)
    _add_hide_personal_data(parser=parser)

    # DEPRECATED in v0.2.0
    # Remove in March 2026
    _add_make_sd_image(parser=parser)

    # Parse arguments
    args = parser.parse_args()

    # Save executed command and arguments into a comment
    comment = _get_command_info(args=args, hide_personal_data=args.hide_personal_data)

    # Handle execution with no arguments
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    # DEPRECATED in v0.2.0
    # Remove in March 2026
    if args.make_sd_image is not None:
        warnings.warn(
            "This argument was deprecated in the release v0.2.0. The spin direction image is now plotted by default, please use --no-html if you want to disable it. This argument will be removed from magnopy in March of 2026.",
            DeprecationWarning,
            stacklevel=2,
        )

    # Process spin values
    if args.spin_values is not None:
        args.spin_values = [float(tmp) for tmp in args.spin_values]

    # Load spin Hamiltonian
    if args.spinham_source.lower() == "tb2j":
        spinham = load_tb2j(
            filename=args.spinham_filename, spin_values=args.spin_values
        )
    elif args.spinham_source.lower() == "grogu":
        spinham = load_grogu(
            filename=args.spinham_filename, spin_values=args.spin_values
        )
    else:
        raise ValueError(
            'Supported sources of spin Hamiltonian are "GROGU" and "TB2J", '
            f'got "{args.spinham_source}".'
        )

    if args.hide_personal_data:
        spinham_filename = args.spinham_filename
    else:
        spinham_filename = os.path.abspath(args.spinham_filename)

    comment += f'\n\nParameters are loaded from "{args.spinham_source.upper()}", source file\n{ICON_IN_FILE} {spinham_filename}'

    optimize_sd(
        spinham=spinham,
        supercell=args.supercell,
        magnetic_field=args.magnetic_field,
        energy_tolerance=args.energy_tolerance,
        torque_tolerance=args.torque_tolerance,
        output_folder=args.output_folder,
        comment=comment,
        no_html=args.no_html,
        hide_personal_data=args.hide_personal_data,
    )



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import numpy as np

from magnopy._spinham._convention import Convention
from magnopy._spinham._hamiltonian import SpinHamiltonian

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def load_grogu(filename, spin_values=None, spglib_types=None) -> SpinHamiltonian:
    r"""
    Reads spin Hamiltonian from the .txt file produced by |GROGU|_.

    For more information on GROGU's file format see |GROGU-FF|_.

    Parameters
    ----------

    filename : str
        Path to the .txt file produced by |GROGU|_.

    spin_values : (M, ) iterable of floats, optional
        Spin values for all magnetic atom. Order is the same as in |GROGU|_ file. Magnetic
        atoms are defined as those that have at least one parameter associated with them.
        If none given, magnopy uses spin values computed from DFT (as provided in the
        |GROGU|_ file).

    spglib_types : (M_prime, ) iterable of ints, optional
        Spglib types for all atoms (not only for magnetic ones, but for all). Order is the
        same as in |GROGU|_ file. If none given, then there will be no "spglib_types" key
        in ``spinham.atoms``.

    Returns
    -------

    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian, that is built from the |GROGU|_ file.

    Raises
    ------

    ValueError
        If ``spin_values`` is provided and its length does not match the number of
        magnetic atoms in the Hamiltonian.

    ValueError
        If ``spglib_types`` is provided and its length does not match the number of
        atoms in the Hamiltonian.

    """

    convention = Convention.get_predefined("grogu")

    # Read the content of the file
    with open(filename, "r", encoding="utf-8") as f:
        lines = f.readlines()

    # Read the cell
    i = 0
    while "cell" not in lines[i].lower() or "(ang)" not in lines[i].lower():
        i += 1

    i += 1

    cell = [
        list(map(float, lines[i].split())),
        list(map(float, lines[i + 1].split())),
        list(map(float, lines[i + 2].split())),
    ]

    # Read the atoms
    while "magnetic" not in lines[i].lower() or "sites" not in lines[i].lower():
        i += 1

    i += 1
    M = int(lines[i].split()[3])
    i += 1

    # Check if spin_values were provided and of correct length
    if spin_values is not None and len(spin_values) != M:
        raise ValueError(f"Expected {M} spin values, got {len(spin_values)}")
    # Check if spglib_types were provided and of correct length
    if spglib_types is not None and len(spglib_types) != M:
        raise ValueError(f"Expected {M} spglib types, got {len(spglib_types)}")

    name_to_index = {}
    atoms = dict(names=[], positions=[], spins=[], g_factors=[2 for _ in range(M)])

    for atom_index in range(M):
        i += 1
        words = lines[i].split()

        name = words[0]
        name_to_index[name] = atom_index

        positions = list(map(float, words[1:4]))
        positions = positions @ np.linalg.inv(cell)

        if spin_values is not None:
            spin = spin_values[atom_index]
        else:
            spin = float(words[4])

        atoms["names"].append(name)
        atoms["positions"].append(positions)
        atoms["spins"].append(spin)

    # Add spglib types if provided
    if spglib_types is not None:
        atoms["spglib_types"] = [int(_) for _ in spglib_types]

    # Construct spin Hamiltonian
    spinham = SpinHamiltonian(convention=convention, cell=cell, atoms=atoms)

    while (
        "intra-atomic" not in lines[i].lower()
        or "anisotropy" not in lines[i].lower()
        or "tensor" not in lines[i].lower()
        or "(mev)" not in lines[i].lower()
    ):
        i += 1

    for _ in range(M):
        i += 2
        name = lines[i].split()[0]
        alpha = name_to_index[name]
        i += 2
        parameter = [
            list(map(float, lines[i].split())),
            list(map(float, lines[i + 1].split())),
            list(map(float, lines[i + 2].split())),
        ]
        i += 2
        spinham.add_21(alpha=alpha, parameter=parameter)

    while (
        "exchange" not in lines[i].lower()
        or "tensor" not in lines[i].lower()
        or "(mev)" not in lines[i].lower()
    ):
        i += 1

    i += 1
    N = int(lines[i].split()[3])
    i += 2

    for _ in range(N):
        i += 2

        words = lines[i].split()
        alpha = name_to_index[words[0]]
        beta = name_to_index[words[1]]
        nu = tuple(list(map(int, words[2:5])))

        i += 2

        parameter = [
            list(map(float, lines[i].split())),
            list(map(float, lines[i + 1].split())),
            list(map(float, lines[i + 2].split())),
        ]

        i += 2

        spinham.add_22(
            alpha=alpha, beta=beta, nu=nu, parameter=parameter, when_present="replace"
        )

    return spinham


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import os

import numpy as np
import wulfric

from magnopy._energy import Energy
from magnopy._lswt import LSWT
from magnopy._package_info import logo
from magnopy._parallelization import multiprocess_over_k
from magnopy.io import plot_dispersion
from magnopy._plotly_engine import PlotlyEngine
from magnopy._constants._icons import ICON_OUT_FILE
from magnopy.io._warnings import _envelope_warning


try:
    import scipy  # noqa F401

    SCIPY_AVAILABLE = True
except ImportError:
    SCIPY_AVAILABLE = False

try:
    import plotly  # noqa F401

    PLOTLY_AVAILABLE = True
except ImportError:
    PLOTLY_AVAILABLE = False

try:
    import matplotlib.pyplot as plt  # noqa F401

    MATPLOTLIB_AVAILABLE = True
except ImportError:
    MATPLOTLIB_AVAILABLE = False


# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def solve_lswt(
    spinham,
    spin_directions=None,
    k_path=None,
    kpoints=None,
    relative=False,
    magnetic_field=None,
    output_folder="magnopy-results",
    number_processors=None,
    comment=None,
    no_html=False,
    hide_personal_data=False,
    spglib_symprec=1e-5,
) -> None:
    r"""
    Computes magnon Hamiltonian at the level of Linear Spin Wave theory.

    Progress of calculation is shown in the standard output (``print()``). A bunch of the
    output files is created and saved on the disk inside the ``output_folder``.

    Parameters
    ----------

    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian object.

    spin_directions : (M, 3) |array-like|_, optional.
        Directions of the local quantization axis for each spin. Magnitude of the vector
        is ignored, only the direction is considered. If ``None``, then magnopy attempts
        to optimize classical energy of spin Hamiltonian to determine spin directions.

    k_path : str, optional
        Specification of the k-path. The format is "G-X-Y|G-Z" For more details
        on the format see documentation of |wulfric|_. If nothing given, then the
        k-path is computed by |wulfric|_ automatically based on the lattice type.
        Ignored if ``kpoints`` are given.

    kpoints : (N, 3) |array-like|_, optional
        Explicit list of k-points to be used instead of automatically generated.

    relative : bool, default False
        If ``relative == True``, then ``kpoints`` are interpreted as given relative to
        the reciprocal unit cell. Otherwise it is interpreted as given in absolute
        coordinates.

    magnetic_field : (3, ) |array-like|_
        Vector of external magnetic field (magnetic flux density, B), given in Tesla.

    output_folder : str, default "magnopy-results"
        Name for the folder where to save the output files. The folder is created if it
        does not exist.

    number_processors : int, optional
        Number of processors to be used in computation. By default magnopy uses all
        available processes. Use ``number_processors=1`` to run in serial mode.

    comment : str, optional
        Any comment, that will be shown in the standard output right after the magnopy's
        logo.

    no_html : bool, default False
        .. versionadded:: 0.2.0

        Whether to produce .html files. If ``no_html = False``, then |plotly|_ is expected
        to be available.

    hide_personal_data : bool, default False
        .. versionadded:: 0.2.0

        If ``False``, then ``os.path.abspath(pathname)`` is used to show full paths to the output
        and input files. If ``True``, then only ``pathname`` is used.

    spglib_symprec : float, default 1e-5
        .. versionadded:: 0.2.0

        Tolerance parameter for the space group symmetry search by |spglib|_. Reduce it
        if the space group is not the one you expected.


    Notes
    -----

    When using this function of magnopy in your Python scripts make sure to safeguard
    your script with the

    .. code-block:: python

        import magnopy

        # Import more stuff
        # or
        # Define your functions, classes

        if __name__ == "__main__":

            # Write your executable code here

    For more information refer to the  "Safe importing of main module" section in
    |multiprocessing|_ docs.

    """

    ################################################################################
    ##                                  Filenames                                 ##
    ################################################################################
    # Create the output directory if it does not exist
    os.makedirs(output_folder, exist_ok=True)

    def envelope_path(pathname):
        if hide_personal_data:
            return pathname
        else:
            return os.path.abspath(pathname)

    SPIN_VECTORS_TXT = envelope_path(os.path.join(output_folder, "SPIN_VECTORS.txt"))
    SPIN_DIRECTIONS_HTML = envelope_path(
        os.path.join(output_folder, "SPIN_DIRECTIONS.html")
    )
    HIGH_SYMMETRY_POINTS_TXT = envelope_path(
        os.path.join(output_folder, "HIGH-SYMMETRY_POINTS.txt")
    )
    K_POINTS_HTML = envelope_path(os.path.join(output_folder, "K-POINTS.html"))
    K_POINTS_TXT = envelope_path(os.path.join(output_folder, "K-POINTS.txt"))
    OMEGAS_TXT = envelope_path(os.path.join(output_folder, "OMEGAS.txt"))
    OMEGAS_PNG = envelope_path(os.path.join(output_folder, "OMEGAS.png"))
    OMEGAS_IMAG_TXT = envelope_path(os.path.join(output_folder, "OMEGAS-IMAG.txt"))
    OMEGAS_IMAG_PNG = envelope_path(os.path.join(output_folder, "OMEGAS-IMAG.png"))
    DELTAS_TXT = envelope_path(os.path.join(output_folder, "DELTAS.txt"))
    DELTAS_PNG = envelope_path(os.path.join(output_folder, "DELTAS.png"))
    E_0_TXT = envelope_path(os.path.join(output_folder, "E_0.txt"))
    E_2_TXT = envelope_path(os.path.join(output_folder, "E_2.txt"))
    ONE_OPERATOR_TERMS_TXT = envelope_path(
        os.path.join(output_folder, "ONE_OPERATOR_TERMS.txt")
    )

    all_good = True

    ################################################################################
    ##                       Logo, comment and plotly check                       ##
    ################################################################################
    # Print logo and a comment
    print(logo(date_time=True, line_length=80))
    if comment is not None:
        print(f"\n{' Comment ':=^80}\n")
        print(comment)

    if no_html:
        print("\nHTML output is disabled by user (no_html=True).")

    if not PLOTLY_AVAILABLE and not no_html:
        print(
            _envelope_warning(
                "Cannot produce files\n  - "
                + "\n  - ".join(
                    [os.path.basename(_) for _ in [K_POINTS_HTML, SPIN_DIRECTIONS_HTML]]
                )
                + "\nbecause plotly is not available.\n"
                "You can install plotly with 'pip install plotly'"
            )
        )

    if not SCIPY_AVAILABLE and not no_html:
        print(
            _envelope_warning(
                "Cannot produce files\n  - "
                + "\n  - ".join([os.path.basename(_) for _ in [K_POINTS_HTML]])
                + "\nbecause scipy is not available."
                "\nYou can install scipy with 'pip install scipy'"
            )
        )

    if not MATPLOTLIB_AVAILABLE:
        print(
            _envelope_warning(
                "Cannot produce files\n  - "
                + "\n  - ".join(
                    [
                        os.path.basename(_)
                        for _ in [OMEGAS_PNG, OMEGAS_IMAG_PNG, DELTAS_PNG]
                    ]
                )
                + "\nbecause matplotlib is not available.\n"
                "You can install matplotlib with 'pip install matplotlib'"
            )
        )

    ################################################################################
    ##                              External effects                              ##
    ################################################################################
    print(f"\n{' External effects ':=^80}\n")

    # Magnetic field
    if magnetic_field is not None:
        print(
            f"Magnetic flux density : "
            f"|{magnetic_field[0]:.5f}, {magnetic_field[1]:.5f}, {magnetic_field[2]:.5f}|"
            f" = {np.linalg.norm(magnetic_field):.5f} Tesla"
        )
        spinham.add_magnetic_field(B=magnetic_field)
    else:
        print("Magnetic flux density : None")
    ################################################################################
    ##                       Optimization of spin directions                      ##
    ################################################################################

    # Energy class will be needed later on as well,
    # thus it is outside of the if-block
    energy = Energy(spinham=spinham)

    if spin_directions is None:
        print(f"\n{' Optimization of spin directions ':=^80}\n")

        print(
            "Spin directions for the ground state are not given, attempt to optimize.\n"
        )

        # Tolerance parameters
        print(f"Energy tolerance      : {1e-5:.5e} meV")
        print(f"Torque tolerance      : {1e-5:.5e}")
        print(
            "Supercell             : 1 x 1 x 1 (original unit cell of the Hamiltonian)"
        )
        print(
            "\nNote: we recommend to obtain ground state outside of the magnopy-lswt program\n"
            "and provide --spin-directions argument to it. See magnopy-optimize-sd, for\n"
            "dedicated spin optimization of magnopy."
        )

        spin_directions = energy.optimize(
            energy_tolerance=1e-5, torque_tolerance=1e-5, quiet=False
        )
    else:
        # Normalize spin directions to unity
        spin_directions = np.array(spin_directions, dtype=float)
        spin_directions = (
            spin_directions / np.linalg.norm(spin_directions, axis=1)[:, np.newaxis]
        )

    ################################################################################
    ##                                Ground state                                ##
    ################################################################################
    print(f"\n{' Ground state ':=^80}\n")

    # Even if they are present in spinham.atoms
    # this function will get the correct ones
    spglib_types = wulfric.get_spglib_types(atoms=spinham.magnetic_atoms)

    name_n = max([4] + [len(name) for name in spinham.magnetic_atoms["names"]])
    print("Order of the atoms and their spglib types are")
    print(f"{'Name':<{name_n}} spglib_type")
    for n_i, name in enumerate(spinham.magnetic_atoms["names"]):
        print(f"{name:{name_n}} {spglib_types[n_i]:^11}")

    # Save spin directions and spin values to the .txt file
    np.savetxt(
        SPIN_VECTORS_TXT,
        np.concatenate(
            (spin_directions, np.array(spinham.magnetic_atoms["spins"])[:, np.newaxis]),
            axis=1,
        ),
        fmt="%12.8f %12.8f %12.8f   %12.8f",
    )
    print(
        f"\nDirections of spin vectors of the ground state and spin values are saved in "
        f"file\n{ICON_OUT_FILE} {SPIN_VECTORS_TXT}"
    )
    # Save spin directions as a .html file
    if not no_html and PLOTLY_AVAILABLE:
        pe = PlotlyEngine()
        pe.plot_cell(
            spinham.cell,
            color="Black",
            legend_label="Unit cell",
        )
        pe.plot_spin_directions(
            positions=np.array(spinham.magnetic_atoms["positions"]) @ spinham.cell,
            spin_directions=spin_directions,
            colors="#A47864",
            legend_label="Spins of the unit cell",
        )
        pe.save(
            output_name=SPIN_DIRECTIONS_HTML,
            axes_visible=True,
            legend_position="top",
            kwargs_write_html=dict(include_plotlyjs=True, full_html=True),
        )
        print(
            f"\nImage of the spin directions is saved in file\n{ICON_OUT_FILE} {SPIN_DIRECTIONS_HTML}"
        )

    # Classical energy
    E_0 = energy.E_0(spin_directions=spin_directions)
    with open(E_0_TXT, "w", encoding="utf-8") as f:
        f.write(f"{E_0:.8f} meV\n")
    print(
        f"\nClassic energy of optimized state (E_0 = {E_0:.3f} meV) is saved in file\n{ICON_OUT_FILE} {E_0_TXT}"
    )

    ################################################################################
    ##                            K-points and k-path                             ##
    ################################################################################
    print(f"\n{' K-points and k-path ':=^80}\n")

    ticks = None
    labels = None
    kpoints_relative = np.array([[]])
    kpoints_absolute = np.array([[]])
    x_data = np.array([])
    if kpoints is not None:
        if relative:
            kpoints_relative = np.array(kpoints, dtype=float)
            kpoints_absolute = kpoints_relative @ wulfric.cell.get_reciprocal(
                cell=spinham.cell
            )
        else:
            kpoints_absolute = np.array(kpoints, dtype=float)
            kpoints_relative = kpoints_absolute @ np.linalg.inv(
                wulfric.cell.get_reciprocal(cell=spinham.cell)
            )

        x_data = np.cumsum(
            np.concatenate(
                (
                    [0.0],
                    np.linalg.norm(
                        kpoints_absolute[1:] - kpoints_absolute[:-1], axis=1
                    ),
                )
            )
        )

        print("K-points are provided by user.")

    else:
        print("Deducing k-points based on the crystal symmetry.")
        print("See wulfric.org for more details on procedure and conventions.")
        spglib_data = wulfric.get_spglib_data(
            cell=spinham.cell, atoms=spinham.atoms, spglib_symprec=spglib_symprec
        )
        print(
            f"\nspglib_symprec  : {spglib_symprec:.5e}.",
            f"Space group     : {spglib_data.space_group_number}",
            f"Bravais lattice : {spglib_data.crystal_family + spglib_data.centring_type}",
            "Convention      : HPKOT",
            sep="\n",
        )
        kp = wulfric.Kpoints.from_crystal(
            cell=spinham.cell,
            atoms=spinham.atoms,
            convention="HPKOT",
            spglib_data=spglib_data,
        )

        # Try to set custom k path
        if k_path is not None:
            try:
                kp.path = k_path
            except ValueError:
                all_good = False
                print(
                    _envelope_warning(
                        "User-provided k-path contains undefined labels of high-symmetry points."
                        "\nPre-defined points are\n  - "
                        + "\n  - ".join(kp.hs_names)
                        + "\nUsing recommended k-path instead.",
                    )
                )

        print(f"K-path          : {kp.path_string}")

        # Save pre-defined high-symmetry points in a .txt file
        label_n = max([5] + [len(label) for label in kp.hs_names])
        with open(HIGH_SYMMETRY_POINTS_TXT, "w", encoding="utf-8") as f:
            f.write(
                f"{'Label':{label_n}} {'k_x':>12} {'k_y':>12} {'k_z':>12}    {'r_b1':>12} {'r_b2':>12} {'r_b3':>12}\n"
            )
            for label in kp.hs_names:
                k_rel = kp.hs_coordinates[label]
                k_abs = k_rel @ kp.rcell
                f.write(
                    f"{label:<{label_n}} {k_abs[0]:12.8f} {k_abs[1]:12.8f} {k_abs[2]:12.8f}    {k_rel[0]:12.8f} {k_rel[1]:12.8f} {k_rel[2]:12.8f}\n"
                )
        print(
            f"\nFull list of pre-defined high-symmetry points is saved in file\n{ICON_OUT_FILE} {HIGH_SYMMETRY_POINTS_TXT}"
        )

        kpoints_relative = kp.points(relative=True)
        kpoints_absolute = kpoints_relative @ kp.rcell
        x_data = kp.flat_points(relative=False)
        ticks = kp.ticks(relative=False)
        labels = kp.labels

        # Produce .html file with the hs points, k-path and brillouin zones
        if not no_html and PLOTLY_AVAILABLE and SCIPY_AVAILABLE:
            pe = wulfric.PlotlyEngine()

            prim_cell, _ = wulfric.crystal.get_primitive(
                cell=spinham.cell,
                atoms=spinham.atoms,
                convention="SC",
                spglib_data=spglib_data,
            )
            pe.plot_brillouin_zone(
                cell=prim_cell,
                color="red",
                legend_label="Brillouin zone of the primitive cell",
            )
            pe.plot_brillouin_zone(
                cell=spinham.cell,
                color="chocolate",
                legend_label="Brillouin zone of the spinham.cell",
            )
            pe.plot_kpath(kp=kp)
            pe.plot_kpoints(kp=kp, only_from_kpath=True)

            pe.save(output_name=K_POINTS_HTML)
            print(
                f"\nHigh-symmetry points and chosen k-path are plotted in\n{ICON_OUT_FILE} {K_POINTS_HTML}"
            )

    # Save k-points info to the .txt file
    np.savetxt(
        K_POINTS_TXT,
        np.concatenate(
            (kpoints_absolute, kpoints_relative, x_data[:, np.newaxis]), axis=1
        ),
        fmt="%12.8f %12.8f %12.8f   %12.8f %12.8f %12.8f   %12.8f",
        header=f"{'k_x':>12} {'k_y':>12} {'k_z':>12}   {'r_b1':>12} {'r_b2':>12} {'r_b3':>12}   {'flat index':>12}",
        comments="",
    )
    print(
        f"\nExplicit list of k-points is saved in file\n{ICON_OUT_FILE} {K_POINTS_TXT}"
    )

    ################################################################################
    ##                                    LSWT                                    ##
    ################################################################################
    print(f"\n{' LSWT ':=^80}\n")
    lswt = LSWT(spinham=spinham, spin_directions=spin_directions)

    # Correction to the classical energy
    E_2 = lswt.E_2()
    with open(E_2_TXT, "w", encoding="utf-8") as f:
        f.write(f"{E_2:.8f} meV\n")
    print(
        f"\nCorrection to the classic ground state energy (E_2 = {E_2:.3f} meV) is saved in file\n{ICON_OUT_FILE} {E_2_TXT}"
    )

    # One-operator coefficients
    one_operator_coefficients = lswt.O()
    np.savetxt(
        ONE_OPERATOR_TERMS_TXT,
        np.concatenate(
            (
                one_operator_coefficients.real[:, np.newaxis],
                one_operator_coefficients.imag[:, np.newaxis],
            ),
            axis=1,
        ),
        fmt="%12.8f %12.8f",
        header=f"{'Re(O_alpha)':>12} {'Im(O_alpha)':>12}",
        comments="",
    )
    print(
        f"\nCoefficients before one-operator term are saved in file\n{ICON_OUT_FILE} {ONE_OPERATOR_TERMS_TXT}"
    )
    print("(shall be zero if the ground state is correct)")

    if not np.allclose(
        one_operator_coefficients, np.zeros(one_operator_coefficients.shape)
    ):
        all_good = False
        print(
            _envelope_warning(
                "Coefficients before the one-operator terms are not zero. It might "
                "indicate that the ground state (spin directions) is not a ground state "
                "of the considered spin Hamiltonian and the results might not be "
                "meaningful.  If O_alpha << 1 then the problem can also be numerical "
                "(due to the finite point arithmetic) and the results are just fine in "
                "that case. Contact developers if you are in doubts: magnopy.org."
            )
        )

    # Compute data for each k-point
    print("\nStart calculations over k-points ... ", end="")
    results = multiprocess_over_k(
        kpoints=kpoints_absolute,
        function=lswt.diagonalize,
        relative=False,
        number_processors=number_processors,
    )
    omegas = np.array([i[0] for i in results]).T
    deltas = np.array([i[1] for i in results])
    n_modes = len(omegas)
    has_imaginary = not np.allclose(omegas.imag, np.zeros(omegas.imag.shape))
    has_nans = np.any(np.isnan(omegas))
    print("Done")

    if has_nans:
        all_good = False
        print(
            _envelope_warning(
                "Some eigenfrequiencies could not be computed (NaN values). It might "
                "indicate that the ground state (spin directions) is not a ground state "
                "of the considered spin Hamiltonian and the results might not be "
                "meaningful. The problem can also be numerical (due to the finite point "
                "arithmetic) and the results are just fine in that case. Contact "
                "developers if you are in doubts: magnopy.org."
            )
        )

    if has_imaginary:
        all_good = False
        print(
            _envelope_warning(
                "Eigenfrequiencies has non-zero imaginary component for some k vectors. "
                "It might indicate that the ground state (spin directions) is not a "
                "ground state of the considered spin Hamiltonian and the results might "
                "not be meaningful. If Im(omega) << 1 then the problem can also be "
                "numerical (due to the finite point arithmetic) and the results are just "
                "fine in that case. Contact developers if you are in doubts: magnopy.org."
            )
        )

    if np.any(omegas.real < -1e-8):
        all_good = False
        print(
            _envelope_warning(
                "Some eigenfrequiencies are negative. This might indicate that the "
                "ground state is not a ground state of the considered spin Hamiltonian "
                "and the results might not be meaningful. Minimum of the spectrum can "
                "indicate a better ground state. Stable ground state should always "
                "have non-negative eigenfrequiencies of the excitations."
            )
        )

    ################################################################################
    ##                                 Text output                                ##
    ################################################################################
    print(f"\n{' Output ':=^80}\n")

    # Omegas
    np.savetxt(
        OMEGAS_TXT,
        omegas.real.T,
        fmt=("%15.6e " * n_modes)[:-1],
        header=" ".join([f"{f'mode {i + 1}':>15}" for i in range(n_modes)]),
        comments="",
    )
    print(f"\nOmegas are saved in file\n{ICON_OUT_FILE} {OMEGAS_TXT}")

    # Deltas
    np.savetxt(DELTAS_TXT, deltas.real, fmt="%10.6e", header="Delta", comments="")
    print(f"Deltas are saved in file\n{ICON_OUT_FILE} {DELTAS_TXT}")

    # Imaginary omegas
    if has_imaginary or has_nans:
        np.savetxt(
            OMEGAS_IMAG_TXT,
            omegas.imag.T,
            fmt=("%15.6e " * n_modes)[:-1],
            header=" ".join([f"{f'mode {i + 1}':>15}" for i in range(n_modes)]),
            comments="",
        )
        print(
            f"Imaginary part of omegas is saved in file\n{ICON_OUT_FILE} {OMEGAS_IMAG_TXT}"
        )

    ################################################################################
    ##                                 png output                                 ##
    ################################################################################

    if MATPLOTLIB_AVAILABLE:
        # Omegas
        plot_dispersion(
            modes=omegas.real,
            x_data=x_data,
            ticks=ticks,
            labels=labels,
            output_filename=OMEGAS_PNG,
            ylabel=R"$\omega_{\alpha}(\boldsymbol{k})$",
        )
        print(f"Plot is saved in file\n{ICON_OUT_FILE} {OMEGAS_PNG}")

        # Deltas
        plot_dispersion(
            modes=deltas.real,
            x_data=x_data,
            ticks=ticks,
            labels=labels,
            output_filename=DELTAS_PNG,
            ylabel=R"$\Delta(\boldsymbol{k})$",
        )
        print(f"Plot is saved in file\n{ICON_OUT_FILE} {DELTAS_PNG}")

        # Imaginary omegas
        if has_imaginary or has_nans:
            plot_dispersion(
                modes=omegas.imag,
                x_data=x_data,
                ticks=ticks,
                labels=labels,
                output_filename=OMEGAS_IMAG_PNG,
                ylabel=R"$\mathcal{Im}(\omega_{\alpha}(\boldsymbol{k}))$",
            )
            print(
                f"Plot of imaginary part is saved in file\n{ICON_OUT_FILE} {OMEGAS_IMAG_PNG}"
            )

    if all_good:
        print(f"\n{' Finished OK ':=^80}")
    else:
        print(f"\n{' Finished with WARNINGS ':=^80}")


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


R"""Exceptions"""

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


class ColpaFailed(Exception):
    r"""
    Raised when diagonalization via Colpa fails.
    """

    def __init__(self):
        message = "Diagonalization via Colpa failed."
        super().__init__(message)


class ConventionError(Exception):
    r"""
    Raised if convention or part of the convention of spin Hamiltonian is not defined.
    """

    def __init__(self, convention, property):
        message = f"Convention of spin Hamiltonian has an undefined property '{property}':\n{convention}"
        super().__init__(message)


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================

r"""
Routines for conversion of the full 3x3 tensor of 22 parameter into its parts and back.
"""

import numpy as np

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def to_iso(parameter, matrix_form=False):
    r"""
    Extracts isotropic part of the full matrix parameter.

    .. math::

        C_{2,2}
        \boldsymbol{S}_{\mu}
        \boldsymbol{J}
        \boldsymbol{S}_{\nu}
        =
        C_{2,2}
        J_{iso}
        \boldsymbol{S}_{\mu}
        \cdot
        \boldsymbol{S}_{\nu}
        +
        C_{2,2}
        \boldsymbol{S}_{\mu}
        \boldsymbol{J}_S
        \boldsymbol{S}_{\nu}
        +
        C_{2,2}
        \boldsymbol{D}
        \cdot
        \left(
        \boldsymbol{S}_{\mu}
        \times
        \boldsymbol{S}_{\nu}
        \right)

    where :math:`J_{iso}` is defined as

    .. math::
        J_{iso} = \dfrac{tr(\boldsymbol{J})}{3}

    This term can be written in the matrix form as

    .. math::

        C_{2,2}
        J_{iso}
        \boldsymbol{S}_{\mu}
        \cdot
        \boldsymbol{S}_{\nu}
        &=
        C_{2,2}
        \boldsymbol{S}_{\mu}
        \boldsymbol{J}_{iso}
        \boldsymbol{S}_{\nu}
        \\
        \boldsymbol{J}_{iso}
        &=
        \begin{pmatrix}
            J_{iso} & 0 & 0 \\
            0 & J_{iso} & 0 \\
            0 & 0 & J_{iso} \\
        \end{pmatrix}

    Parameters
    ----------

    parameter : (3, 3) |array-like|_
        Full matrix of the exchange parameter (:math:`\boldsymbol{J}`).

    matrix_form : bool, default False
        Whether to return isotropic part of the matrix instead of isotropic parameter.

    Returns
    -------

    iso : float or (3, 3) :numpy:`ndarray`
        Isotropic parameter.

        * If ``matrix_form == False``, then returns a number :math:`J_{iso}`.
        * If ``matrix_form == True``, then returns a matrix :math:`\boldsymbol{J}_{iso}`.

    See Also
    --------

    from_iso
    to_dmi
    to_symm_anisotropy

    Examples
    --------

    .. doctest::

        >>> from magnopy import converter22
        >>> matrix = [[1, 3, 4], [-1, -2, 3], [4, 0, 10]]
        >>> converter22.to_iso(matrix)
        3.0
        >>> converter22.to_iso(matrix, matrix_form=True)
        array([[3., 0., 0.],
               [0., 3., 0.],
               [0., 0., 3.]])
    """

    iso = (parameter[0][0] + parameter[1][1] + parameter[2][2]) / 3

    if matrix_form:
        return iso * np.eye(3, dtype=float)

    return float(iso)


def to_symm_anisotropy(parameter):
    r"""
    Extracts traceless, symmetric anisotropic part of the full matrix parameter.

    .. math::

        C_{2,2}
        \boldsymbol{S}_{\mu}
        \boldsymbol{J}
        \boldsymbol{S}_{\nu}
        =
        C_{2,2}
        J_{iso}
        \boldsymbol{S}_{\mu}
        \cdot
        \boldsymbol{S}_{\nu}
        +
        C_{2,2}
        \boldsymbol{S}_{\mu}
        \boldsymbol{J}_S
        \boldsymbol{S}_{\nu}
        +
        C_{2,2}
        \boldsymbol{D}
        \cdot
        \left(
        \boldsymbol{S}_{\mu}
        \times
        \boldsymbol{S}_{\nu}
        \right)

    where matrix :math:`\boldsymbol{J}_S` is defined as

    .. math::

        \boldsymbol{J}_S
        =
        \dfrac{\boldsymbol{J}
        +
        \boldsymbol{J}^T}{2}

    Parameters
    ----------

    parameter : (3, 3) |array-like|_
        Full matrix of the exchange parameter (:math:`\boldsymbol{J}`).

    Returns
    -------

    aniso : float or (3, 3) :numpy:`ndarray`
        Matrix of a traceless, symmetric anisotropy.

    See Also
    --------

    to_iso
    to_dmi

    Examples
    --------

    .. doctest::

        >>> import magnopy
        >>> matrix = [[1, 3, 4], [-1, -2, 0], [4, 0, 10]]
        >>> magnopy.converter22.to_symm_anisotropy(matrix)
        array([[-2.,  1.,  4.],
               [ 1., -5.,  0.],
               [ 4.,  0.,  7.]])
    """

    parameter = np.array(parameter)

    return (parameter + parameter.T) / 2 - to_iso(parameter=parameter, matrix_form=True)


def to_dmi(parameter, matrix_form=False):
    r"""
    Extracts antisymmetric part (DMI) of the full matrix parameter.


    .. math::

        C_{2,2}
        \boldsymbol{S}_{\mu}
        \boldsymbol{J}
        \boldsymbol{S}_{\nu}
        =
        C_{2,2}
        J_{iso}
        \boldsymbol{S}_{\mu}
        \cdot
        \boldsymbol{S}_{\nu}
        +
        C_{2,2}
        \boldsymbol{S}_{\mu}
        \boldsymbol{J}_S
        \boldsymbol{S}_{\nu}
        +
        C_{2,2}
        \boldsymbol{D}
        \cdot
        \left(
        \boldsymbol{S}_{\mu}
        \times
        \boldsymbol{S}_{\nu}
        \right)

    where vector :math:`\boldsymbol{D}` is defined as

    .. math::

        \begin{pmatrix}
            0 & D^z & -D^y \\
            -D^z & 0 & D^x \\
            D^y & -D^x & 0 \\
        \end{pmatrix}
        =
        \dfrac{\boldsymbol{J}
        -
        \boldsymbol{J}^T}{2}
        =
        \boldsymbol{J}_A


    Parameters
    ----------

    parameter : (3, 3) |array-like|_
        Full matrix of the exchange parameter (:math:`\boldsymbol{J}`).

    matrix_form : bool, default False
        Whether to return dmi in a matrix form instead of a vector.

    Returns
    -------

    dmi : (3,) or (3, 3) :numpy:`ndarray`
        Antisymmetric exchange (DMI).

        * If ``matrix_form == False``, then returns a vector :math:`\boldsymbol{D}`.
        * If ``matrix_form == True``, then returns a matrix :math:`\boldsymbol{J}_A`.

    See Also
    --------

    from_dmi
    to_iso
    to_symm_anisotropy

    Examples
    --------

    .. doctest::

        >>> from magnopy import converter22
        >>> parameter = [[1, 3, 0], [-1, -2, 3], [0, 3, 9]]
        >>> converter22.to_dmi(parameter)
        array([0., 0., 2.])
        >>> converter22.to_dmi(parameter, matrix_form = True)
        array([[ 0.,  2.,  0.],
               [-2.,  0.,  0.],
               [ 0.,  0.,  0.]])
    """

    parameter = np.array(parameter)

    asymm_matrix = (parameter - parameter.T) / 2

    if matrix_form:
        return asymm_matrix

    return np.array(
        [asymm_matrix[1][2], asymm_matrix[2][0], asymm_matrix[0][1]],
        dtype=float,
    )


def from_iso(iso):
    r"""
    Computes matrix form of the isotropic exchange parameter.

    .. math::

        C_{2,2}
        J_{iso}
        \boldsymbol{S}_{\mu}
        \cdot
        \boldsymbol{S}_{\nu}
        =
        C_{2,2}
        \boldsymbol{S}_{\mu}
        \boldsymbol{J}_{iso}
        \boldsymbol{S}_{\nu}

    where matrix :math:`\boldsymbol{J}_{iso}` is defined as

    .. math::

        \boldsymbol{J}_{iso}
        =
        \begin{pmatrix}
            J_{iso} & 0 & 0 \\
            0 & J_{iso} & 0 \\
            0 & 0 & J_{iso} \\
        \end{pmatrix}

    Parameters
    ----------

    iso : int or float
        Isotropic exchange parameter.

    Returns
    -------

    parameter : (3, 3) :numpy:`ndarray`
        Matrix form of the isotropic exchange parameter (:math:`\boldsymbol{J}_{iso}`).

    See Also
    --------

    to_iso
    from_dmi

    Examples
    --------

    .. doctest::

        >>> from magnopy import converter22
        >>> converter22.from_iso(iso=1)
        array([[1., 0., 0.],
               [0., 1., 0.],
               [0., 0., 1.]])
    """

    return iso * np.eye(3, dtype=float)


def from_dmi(dmi):
    r"""
    Computes matrix form of the Dzyaloshinskii-Moriya interaction parameter.

    .. math::


        C_{2,2}
        \boldsymbol{D}
        \cdot
        \left(
        \boldsymbol{S}_{\mu}
        \times
        \boldsymbol{S}_{\nu}
        \right)
        =
        C_{2,2}
        \boldsymbol{S}_{\mu}
        \boldsymbol{J}_A
        \boldsymbol{S}_{\nu}


    where matrix :math:`\boldsymbol{J}_A` is defined as

    .. math::

        \boldsymbol{J}_A
        =
        \begin{pmatrix}
            0 & D^z & -D^y \\
            -D^z & 0 & D^x \\
            D^y & -D^x & 0 \\
        \end{pmatrix}

    Parameters
    ----------

    dmi : (3,) |array-like|_
        Vector of Dzyaloshinskii-Moriya interaction parameter :math:`(D_x, D_y, D_z)`.

    Returns
    -------

    parameter : (3, 3) :numpy:`ndarray`
        Matrix form of the Dzyaloshinskii-Moriya interaction parameter
        (:math:`\boldsymbol{J}_A`).

    See Also
    --------

    to_dmi
    from_iso


    Examples
    --------

    .. doctest::

        >>> from magnopy import converter22
        >>> converter22.from_dmi(dmi = (1, 2, 0))
        array([[ 0.,  0., -2.],
               [ 0.,  0.,  1.],
               [ 2., -1.,  0.]])
    """

    return np.array(
        [
            [0, dmi[2], -dmi[1]],
            [-dmi[2], 0, dmi[0]],
            [dmi[1], -dmi[0], 0],
        ],
        dtype=float,
    )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================

import warnings
import numpy as np


try:
    import matplotlib.pyplot as plt

    MATPLOTLIB_AVAILABLE = True
except ImportError:
    MATPLOTLIB_AVAILABLE = False
    MATPLOTLIB_WARNING_MESSAGE = "Matplotlib is not available. Please install 'magnopy[visual]' or matplotlib itself."

from magnopy._constants._units import _MAGNON_ENERGY_UNITS


def plot_dispersion(
    modes,
    x_data=None,
    ticks=None,
    labels=None,
    output_filename=None,
    ylabel="Energy",
    colors=None,
) -> None:
    r"""
    Plot dispersion curves.

    Input data (``modes``) are expected in meV. meV converted into THz for the twin axis
    via

    .. math::
        E = h \nu

    Parameters
    ----------

    modes : (M,) or (N, M) |array-like|_
        Dispersion data, Either one or :math:`N` modes with :math:`M` points each.

    x_data : (M,) |array-like|_, optional
        Abscissa (x axis) data. If ``None``, then integer indices would
        be used.

    ticks : (K,) |array-like|_, optional
        Positions of high-symmetry points along the x axis. If ``None``, then
        no high-symmetry points would be marked. coordinates are interpreted
        together with ``x_data``, whether the latter is provided or not.

    labels : (K,) list of str, optional
        Labels of high-symmetry points.

    output_filename : str, optional
        Name of the file for saving the image. If ``None``, then the graph would be
        opened in the interactive matplotlib window.

    ylabel : str, default "Energy"
        Label for the ordinate (y axis). Do not include units, units are included
        automatically - meV for primary axis and THz for twin axis.

    colors : "random" or (N,) |array-like|_, optional
        Colors for each mode. If "random", random colors are assigned. If ``None``, then a
        default color is used for all modes.

    Raises
    ------

    ValueError
        If shapes of ``x_data`` and ``modes`` are incompatible, i. e.
        ``x_data.shape[0] != modes.shape[1]``.

    ValueError
        If lengths of ``ticks`` and ``labels`` are incompatible, i. e.
        ``len(ticks) != len(labels)``.

    ValueError
        If length of ``colors`` is incompatible with number of modes, i. e.
        ``len(colors) != modes.shape[0]``.

    Notes
    -----

    If ``matplotlib`` is not available, a warning is issued and the function plots
    nothing.

    * If ``ticks`` is given and ``labels`` is not, labels are filled as ``K1, K2, ...``.
    * If ``labels`` is given and ``ticks`` is not, ``labels`` is ignored.
    * If both ``ticks`` and ``labels`` are not given, no high-symmetry points are marked.
    """

    if not MATPLOTLIB_AVAILABLE:
        warnings.warn(MATPLOTLIB_WARNING_MESSAGE, RuntimeWarning, stacklevel=2)
        return

    modes = np.array(modes)

    if len(modes.shape) == 1:
        modes = modes[np.newaxis, :]

    if x_data is None:
        x_data = np.arange(modes.shape[1])
    else:
        x_data = np.array(x_data)

    if x_data.shape[0] != modes.shape[1]:
        raise ValueError(
            f"Incompatible shapes between x_data ({x_data.shape}) and modes ({modes.shape}): {x_data.shape[0]} != {modes.shape[1]}."
        )

    if ticks is not None and labels is None:
        labels = [f"K{i + 1}" for i in range(len(ticks))]

    if ticks is not None and len(labels) != len(ticks):
        raise ValueError(
            f"Incompatible lengths between ticks ({len(ticks)}) and labels ({len(labels)}): {len(ticks)} != {len(labels)}."
        )

    if colors is None:
        colors = ["#A47864" for _ in range(modes.shape[0])]
    elif colors == "random":
        import random

        colors = []
        for _ in range(modes.shape[0]):
            color = "#"
            for _ in range(6):
                color += random.choice("0123456789ABCDEF")
            colors.append(color)
    elif len(colors) != modes.shape[0]:
        raise ValueError(
            f"Incompatible length of colors ({len(colors)}) and number of modes ({modes.shape[0]}): {len(colors)} != {modes.shape[0]}."
        )

    fig, ax = plt.subplots()

    for mode, color in zip(modes, colors):
        ax.plot(x_data, mode, lw=1, color=color)

    if ticks is not None:
        ax.set_xticks(ticks=ticks, labels=labels, fontsize=13)
        ax.vlines(
            ticks,
            0,
            1,
            lw=0.5,
            color="grey",
            ls="dashed",
            zorder=0,
            transform=ax.get_xaxis_transform(),
        )
    else:
        ax.set_xticks([], [])
        ax.set_xlabel("k-points (arbitrary units)", fontsize=15)

    ax.set_xlim(x_data[0], x_data[-1])
    ax.set_ylabel(f"{ylabel}, meV", fontsize=15)

    ax.hlines(
        0,
        0,
        1,
        lw=0.5,
        color="grey",
        linestyle="dashed",
        transform=ax.get_yaxis_transform(),
    )

    # Add twin axis

    ylims = ax.get_ylim()

    meV_to_THz = _MAGNON_ENERGY_UNITS["mev"] / _MAGNON_ENERGY_UNITS["thz"]

    twinax = ax.twinx()

    twinax.set_ylim(meV_to_THz * ylims[0], meV_to_THz * ylims[1])

    twinax.set_ylabel(f"{ylabel}, THz", fontsize=15)

    if output_filename is not None:
        plt.savefig(output_filename, dpi=400, bbox_inches="tight")
    else:
        plt.show()

    plt.close()



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


from multiprocessing import Pool

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def multiprocess_over_k(
    kpoints, function, relative=False, units="meV", number_processors=None
):
    r"""
    Parallelizes calculation over the kpoints using |multiprocessing|_ module.

    Parameters
    ----------

    kpoints : (N, 3) |array-like|_
        List of the kpoints.

    function : callable
        Function that process one kpoint and is called as

        .. code-block:: python

            result = function(kpoints[i], relative, units)

    relative : bool, default False
        If ``relative=True``, then ``k`` is interpreted as given relative to the
        reciprocal unit cell. Otherwise it is interpreted as given in absolute
        coordinates.

    units : str, default "meV"
        .. versionadded:: 0.3.0

        Units of energy. See :py:attr:`.SpinHamiltonian.units` for the list of
        supported units.

    number_processors : int, optional
        By default magnopy uses all available processes. Pass ``number_processors=1`` to
        run in serial.

    Returns
    -------

    results : (N, ) list
        List of objects that are returned by the ``function``.

    Notes
    -----

    When using this function of magnopy in your Python scripts make sure to safeguard
    your script with the

    .. code-block:: python

        import magnopy

        # Import more stuff
        # or
        # Define your functions, classes

        if __name__ == "__main__":

            # Write your executable code here

    For more information refer to the  "Safe importing of main module" section in
    |multiprocessing|_ docs.
    """

    relative = [relative for _ in kpoints]
    units = [units for _ in kpoints]

    if number_processors == 1:
        results = list(map(function, kpoints, relative, units))
    else:
        with Pool(number_processors) as p:
            results = p.starmap(function, zip(kpoints, relative, units))

    return results


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


from calendar import month_name
from datetime import datetime

from magnopy import __doclink__, __release_date__, __version__

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")

# fmt: off
BINARY_LOGO  =  [
    [1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    [1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    [1,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    [1,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    [1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    [0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    [0,1,1,1,0,0,0,1,0,0,0,1,0,0,1,1,1,0,0,0,1,1,1,0,0,1,0,0,0,1,0,0,1,1,1,0,0,1,1,1,1,0,0,1,0,0,0,1],
    [0,1,1,1,1,0,0,1,1,0,1,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,1,0,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,1,1,0,1,1],
    [0,1,1,1,1,0,0,1,0,1,0,1,0,1,1,1,1,1,0,1,0,0,1,1,0,1,0,1,0,1,0,1,0,0,0,1,0,1,1,1,1,0,0,0,1,1,1,0],
    [0,1,1,1,1,0,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,1,1,0,1,0,0,0,1,0,1,0,0,0,0,0,0,0,1,0,0],
    [0,1,1,1,1,0,0,1,0,0,0,1,0,1,0,0,0,1,0,0,1,1,1,0,0,1,0,0,0,1,0,0,1,1,1,0,0,1,0,0,0,0,0,0,0,1,0,0],
    [0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0],
]
# fmt: on


def logo(info=None, line_length=None, flat=False, date_time=False, comment=None):
    """
    Generates a logo of Magnopy.

    Returns the logo and some information about the package.

    Parameters
    ----------

    info : list of str, optional
        Information about the package, that is displayed below the logo. Each element
        should not exceed 58 characters. By default it displays the version, link to the
        documentation, release date and license. Pass an empty list to display the logo
        only.

    line_length : int, optional
        Length of the lines to be returned. Minimum value is 70.

    flat : bool
        Whether to return a flat logo or not.

    date_time : bool, default False
        Whether to include the date and time to the info or not.

    comment : str or bool, optional
        Whether to use some character at the beginning of each string. If ``bool`` and
        ``True``, then "# " is used. If ``str``, then this string is used. If ``None``,
        then no character is used.

    Returns
    -------

    logo_info : str
        Logo and information about the package.

    """
    if info is None:
        info = [
            f"Version: {__version__}",
            f"Documentation: {__doclink__}",
            f"Release date: {__release_date__}",
            "License: GNU GPLv3",
            f"Copyright (C) 2023-{datetime.now().year}  Magnopy Team",
        ]
    if date_time:
        cd = datetime.now()
        info.append("")
        info.append(
            f"Generated on {cd.day} {month_name[cd.month]} {cd.year}"
            + f" at {cd.hour}:{cd.minute}:{cd.second} "
        )
    logo = [
        "███╗   ███╗  █████╗   ██████╗  ███╗   ██╗  ██████╗  ██████╗  ██╗   ██╗",
        "████╗ ████║ ██╔══██╗ ██╔════╝  ████╗  ██║ ██╔═══██╗ ██╔══██╗ ╚██╗ ██╔╝",
        "██╔████╔██║ ███████║ ██║  ███╗ ██╔██╗ ██║ ██║   ██║ ██████╔╝  ╚████╔╝ ",
        "██║╚██╔╝██║ ██╔══██║ ██║  ╚██║ ██║╚██╗██║ ██║   ██║ ██╔═══╝    ╚██╔╝  ",
        "██║ ╚═╝ ██║ ██║  ██║ ╚██████╔╝ ██║ ╚████║ ╚██████╔╝ ██║         ██║   ",
        "╚═╝     ╚═╝ ╚═╝  ╚═╝  ╚═════╝  ╚═╝  ╚═══╝  ╚═════╝  ╚═╝         ╚═╝   ",
    ]
    if flat:
        logo = [
            "███    ███   █████    ██████   ███    ██   ██████   ██████   ██    ██",
            "████  ████  ██   ██  ██        ████   ██  ██    ██  ██   ██   ██  ██ ",
            "██ ████ ██  ███████  ██   ███  ██ ██  ██  ██    ██  ██████     ████  ",
            "██  ██  ██  ██   ██  ██    ██  ██  ██ ██  ██    ██  ██          ██   ",
            "██      ██  ██   ██   ██████   ██   ████   ██████   ██          ██   ",
        ]
    cat = [
        "▄   ▄     ",
        "█▀█▀█     ",
        "█▄█▄█     ",
        " ███   ▄▄ ",
        " ████ █  █",
        " ████    █",
        " ▀▀▀▀▀▀▀▀ ",
    ]

    N = len(logo[0])
    n = len(cat[0]) + 2
    if line_length is None:
        line_length = N
    if line_length < N:
        line_length = N
    if isinstance(comment, bool) and comment:
        comment = "# "
    elif comment is not None:
        comment = str(comment)
    else:
        comment = ""

    logo_info = [f"{x:^{N}}" for x in logo]
    if len(info) > 0:
        if len(info) <= len(cat):
            before = (len(cat) - len(info)) // 2 + (len(cat) - len(info)) % 2
            after = len(cat) - len(info) - before
            for i in range(len(cat)):
                if i < before or i >= len(cat) - after:
                    logo_info.append(f"{' ':{N - n}}{cat[i]:^{n}}")
                else:
                    logo_info.append(f"{info[i - before]:^{N - n}}{cat[i]:^{n}}")
        else:
            before = (len(info) - len(cat)) // 2
            after = len(info) - len(cat) - before
            for i in range(len(info)):
                if i < before or i >= len(info) - after:
                    logo_info.append(f"{info[i - before]:^{N - n}}")
                else:
                    logo_info.append(
                        f"{info[i - before]:^{N - n}}{cat[i - before]:^{n}}"
                    )

    logo_info = [f"{comment}{x:^{line_length}}\n" for x in logo_info]
    return "".join(logo_info)[:-1]


def _warranty():
    r"""
    Outputs short warranty summary for terminal interactions
    """

    return """THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
ALL NECESSARY SERVICING, REPAIR OR CORRECTION."""


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import numpy as np

from magnopy._data_validation import (
    _spins_ordered,
    _validate_atom_index,
    _validate_unit_cell_index,
    _validated_units,
)
from magnopy._constants._units import _PARAMETER_UNITS


def _get_primary_p32(alpha, beta, nu, parameter=None, S_alpha=None, S_beta=None):
    r"""
    Return the primary version of the parameter.

    For the definition of the primary version see
    :ref:`user-guide_theory-behind_multiple-counting`.

    Parameters
    ----------
    alpha : int
        Index of the first atom.

    beta : int
        Index of the second atom.

    nu : tuple of 3 int
        Unit cell for the second atom.

    parameter : (3, 3, 3) :numpy:`ndarray`, optional
        Full matrix of the parameter.

    S_alpha : float, optional
        Spin value of atom ``alpha``

    S_beta : float, optional
        Spin value of atom ``beta``

    Returns
    -------

    alpha : int
        Index of the first atom.

    beta : int
        Index of the second atom.

    nu : tuple of 3 int
        Unit cell for the second atom.

    parameter : (3, 3, 3) :numpy:`ndarray`
        Full matrix of the parameter. It is returned only if ``parameter is not None``.
    """

    if _spins_ordered(mu1=(0, 0, 0), alpha1=alpha, mu2=nu, alpha2=beta):
        pass
    else:
        i, j, k = nu
        alpha, beta, nu = beta, alpha, (-i, -j, -k)
        if parameter is not None:
            parameter = np.transpose(parameter, (2, 1, 0)) * S_alpha / S_beta

    if parameter is None:
        return alpha, beta, nu
    return alpha, beta, nu, parameter


class _P32_iterator:
    R"""
    Iterator over the (three spins & two sites) parameters of the spin Hamiltonian.
    """

    def __init__(self, spinham) -> None:
        self.container = spinham._32
        self.mc = spinham.convention.multiple_counting
        self.length = len(self.container)
        self.index = 0
        self.spins = spinham.atoms.spins

    def __next__(self):
        if self.index < self.length:
            self.index += 1
            return self.container[self.index - 1]

        elif self.mc and self.index < 2 * self.length:
            self.index += 1
            alpha, beta, (i, j, k), parameter = self.container[
                self.index - 1 - self.length
            ]
            return [
                beta,
                alpha,
                (-i, -j, -k),
                np.transpose(parameter, (2, 1, 0))
                * self.spins[alpha]
                / self.spins[beta],
            ]

        raise StopIteration

    def __len__(self):
        return self.length * (1 + int(self.mc))

    def __iter__(self):
        return self


@property
def _p32(spinham):
    r"""
    Parameters of (three spins & two sites) term of the Hamiltonian.

    .. math::

        \boldsymbol{J}_{3,2}(\boldsymbol{r}_{\nu,\alpha\beta})

    of the term

    .. math::

        C_{3,2}
        \sum_{\substack{\mu, \nu, \alpha, \beta,\\ i, j, u}}
        J^{iju}_{3,2}(\boldsymbol{r}_{\nu,\alpha\beta})
        S_{\mu,\alpha}^i
        S_{\mu,\alpha}^j
        S_{\mu+\nu,\beta}^u

    Returns
    -------

    parameters : iterator
        List of parameters. The list has a form of

        .. code-block:: python

            [[alpha, beta, nu, J], ...]

        where

        ``alpha`` is an index of the atom located in the (0,0,0) unit cell.

        ``beta`` is an index of the atom located in the  nu unit cell.

        ``nu`` defines the unit cell of the second atom (beta). It is a tuple of 3
        integers.

        ``J`` is a (3, 3, 3) :numpy:`ndarray`.

    See Also
    --------

    add_32
    remove_32
    """

    return _P32_iterator(spinham)


# ARGUMENT "replace" DEPRECATED since 0.4.0
# Remove in May of 2026
def _add_32(
    spinham,
    alpha: int,
    beta: int,
    nu: tuple,
    parameter,
    units=None,
    when_present="raise error",
    replace=None,
) -> None:
    r"""
    Adds a (three spins & two sites) parameter to the Hamiltonian.

    Doubles of the bonds are managed automatically (independently of the convention of the
    Hamiltonian).


    Parameters
    ----------
    alpha : int
        Index of an atom from the (0, 0, 0) unit cell.

        ``0 <= alpha < len(spinham.atoms.names)``.

    beta : int
        Index of an atom from the nu unit cell.

        ``0 <= beta < len(spinham.atoms.names)``.

    nu : tuple of 3 int
        Three relative coordinates with respect to the three lattice vectors, that
        specify the unit cell for the second atom.

        .. math::

            \nu
            =
            (x_{\boldsymbol{a}_1}, x_{\boldsymbol{a}_2}, x_{\boldsymbol{a}_3})

    parameter : (3, 3, 3) |array-like|_
        Value of the parameter (:math:`3\times3\times3` matrix). Given in the units of ``units``.

    units : str, optional
        .. versionadded:: 0.3.0

        Units in which the ``parameter`` is given. Parameters have the the units of energy.
        By default assumes :py:attr:`.SpinHamiltonian.units`. For the list of the supported
        units see :ref:`user-guide_usage_units_parameter-units`. If given ``units`` are different from
        :py:attr:`.SpinHamiltonian.units`, then the parameter's value will be converted
        automatically from ``units`` to :py:attr:`.SpinHamiltonian.units`.

    when_present : str, default "raise error"
        .. versionadded:: 0.4.0

        Action to take if a pair of atoms already has a parameter associated with it.
        Case-insensitive. Supported values are:

        - ``"raise error"`` (default): raises an error if a pair of atoms already has a
          parameter associated with it.
        - ``"replace"``: replace existing value of the parameter with the new one.
        - ``"add"``: add the value of the parameter to the existing one.
        - ``"mean"``: replace the value of the parameter with the arithmetic mean of
          existing and new parameters.
        - ``"skip"``: Leave existing parameter unchanged and continue without raising an
          error.

    replace : bool, default False
        Whether to replace the value of the parameter if a pair of atoms already has a
        parameter associated with it.

        .. deprecated:: 0.4.0
            The ``replace`` argument will be removed in May of 2026. Use
            ``modify="replace"`` instead.


    Raises
    ------

    ValueError
        If a pair of atoms already has a parameter associated with it and ``when_present="raise error"``.

    ValueError
        If ``when_present`` has an unsupported value.

    See Also
    --------

    p32
    remove_32

    Notes
    -----

    If ``spinham.convention.multiple_counting`` is ``True``, then this function adds both
    the bond and its double to the Hamiltonian. It will cause an ``ValueError`` to
    add the double of the bond after the bond is added.

    If ``spinham.convention.multiple_counting`` is ``False``, then only the primary
    version of the bond is added to the Hamiltonian.

    For the definition of the primary version see
    :ref:`user-guide_theory-behind_multiple-counting`.
    """

    if replace is not None:
        import warnings

        warnings.warn(
            'The "replace" argument is deprecated since version 0.4.0 and will be removed in May of 2026. Use when_present="replace" instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        if replace:
            when_present = "replace"
        else:
            when_present = "raise error"

    _validate_atom_index(index=alpha, atoms=spinham.atoms)
    _validate_atom_index(index=beta, atoms=spinham.atoms)
    _validate_unit_cell_index(ijk=nu)
    spinham._reset_internals()

    parameter = np.array(parameter)

    if units is not None:
        units = _validated_units(units=units, supported_units=_PARAMETER_UNITS)
        parameter = (
            parameter * _PARAMETER_UNITS[units] / _PARAMETER_UNITS[spinham._units]
        )

    alpha, beta, nu, parameter = _get_primary_p32(
        alpha=alpha,
        beta=beta,
        nu=nu,
        parameter=parameter,
        S_alpha=spinham.atoms.spins[alpha],
        S_beta=spinham.atoms.spins[beta],
    )

    # TODO BINARY SEARCH
    # Try to find the place for the new one inside the list
    index = 0
    while index < len(spinham._32):
        # If already present in the model
        if spinham._32[index][:3] == [alpha, beta, nu]:
            # Either replace
            if when_present.lower() == "replace":
                spinham._32[index][3] = parameter
            # Or add
            elif when_present.lower() == "add":
                spinham._32[index][3] = spinham._32[index][3] + parameter
            # Or replace with mean value
            elif when_present.lower() == "mean":
                spinham._32[index][3] = (spinham._32[index][3] + parameter) / 2.0
            # Or do nothing
            elif when_present.lower() == "skip":
                pass
            # Or raise an error
            elif when_present.lower() == "raise error":
                raise ValueError(
                    f"(Three spins & two sites) parameter is already set for the pair of atoms {alpha} and {beta} ({nu}). Or for their double bond."
                )
            else:
                raise ValueError(
                    f'Unsupported value of when_present: "{when_present}". Supported values are: "raise error", "replace", "add", "mean", "skip".'
                )

            return

        # If it should be inserted before current element
        if spinham._32[index][:3] > [alpha, beta, nu]:
            spinham._32.insert(index, [alpha, beta, nu, parameter])
            return

        index += 1

    # If it should be inserted at the end or at the beginning of the list
    spinham._32.append([alpha, beta, nu, parameter])


def _remove_32(spinham, alpha: int, beta: int, nu: tuple) -> None:
    r"""
    Removes a (three spins & two sites) parameter from the Hamiltonian.

    Doubles of the bonds are managed automatically (independently of the convention of the
    Hamiltonian).

    Parameters
    ----------

    alpha : int
        Index of an atom from the (0, 0, 0) unit cell.

        ``0 <= alpha < len(spinham.atoms.names)``.

    beta : int
        Index of an atom from the nu unit cell.

        ``0 <= beta < len(spinham.atoms.names)``.

    nu : tuple of 3 int
        Three relative coordinates with respect to the three lattice vectors, that
        specify the unit cell for the second atom.

        .. math::

            \nu
            =
            (x_{\boldsymbol{a}_1}, x_{\boldsymbol{a}_2}, x_{\boldsymbol{a}_3})

    See Also
    --------

    p32
    add_32

    Notes
    -----

    If ``spinham.convention.multiple_counting`` is ``True``, then this function removes
    all versions of the bond from the Hamiltonian.

    If ``spinham.convention.multiple_counting`` is ``False``, then this function removes
    the primary version of the given bond.

    For the definition of the primary version see
    :ref:`user-guide_theory-behind_multiple-counting`.

    For instance, if ``(1, 0, (0, 0, 0))`` is given, then this function attempts to
    remove either both ``(1, 0, (0, 0, 0))`` and ``(0, 1, (0, 0, 0))`` if
    ``spinham.convention.multiple_counting == True`` or the primary version
    ``(0, 1, (0, 0, 0))`` if ``spinham.convention.multiple_counting == False``.
    """

    _validate_atom_index(index=alpha, atoms=spinham.atoms)
    _validate_atom_index(index=beta, atoms=spinham.atoms)
    _validate_unit_cell_index(ijk=nu)

    alpha, beta, nu = _get_primary_p32(alpha=alpha, beta=beta, nu=nu)

    # TD-BINARY_SEARCH

    for index in range(len(spinham._32)):
        # As the list is sorted, there is no point in resuming the search
        # when a larger element is found
        if spinham._32[index][:3] > [alpha, beta, nu]:
            return

        if spinham._32[index][:3] == [alpha, beta, nu]:
            del spinham._32[index]
            spinham._reset_internals()
            return



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import numpy as np

from magnopy._data_validation import _validate_atom_index, _validated_units
from magnopy._constants._units import _PARAMETER_UNITS


@property
def _p31(spinham) -> list:
    r"""
    Parameters of (three spins & one site) term of the Hamiltonian.

    .. math::

        \boldsymbol{J}_{3,1}(\boldsymbol{r}_{\alpha})

    of the term

    .. math::

        C_{3,1}
        \sum_{\mu, \alpha, i, j, u}
        J^{iju}_{3,1}(\boldsymbol{r}_{\alpha})
        S_{\mu,\alpha}^i
        S_{\mu,\alpha}^j
        S_{\mu,\alpha}^u

    Returns
    -------

    parameters : list
        List of parameters. The list has a form of

        .. code-block:: python

            [[alpha, J], ...]

        ``0 <= len(parameters) <= len(spinham.atoms.names)``.

        where ``alpha`` is an index of the atom to which the parameter is assigned and
        ``J`` is a (3, 3, 3) :numpy:`ndarray`. The parameters are sorted by the index of
        an atom ``alpha`` in the ascending order.

    See Also
    --------

    add_31
    remove_31
    """

    return spinham._31


# ARGUMENT "replace" DEPRECATED since 0.4.0
# Remove in May of 2026
def _add_31(
    spinham, alpha: int, parameter, units=None, when_present="raise error", replace=None
) -> None:
    r"""
    Adds a (three spins & one site) parameter to the Hamiltonian.

    Parameters
    ----------

    alpha : int
        Index of an atom, with which the parameter is associated.

        ``0 <= alpha < len(spinham.atoms.names)``.

    parameter : (3, 3, 3) |array-like|_
        Value of the parameter (:math:`3\times3\times3` tensor). Given in the units of ``units``.

    units : str, optional
        .. versionadded:: 0.3.0

        Units in which the ``parameter`` is given. Parameters have the the units of energy.
        By default assumes :py:attr:`.SpinHamiltonian.units`. For the list of the supported
        units see :ref:`user-guide_usage_units_parameter-units`. If given ``units`` are different from
        :py:attr:`.SpinHamiltonian.units`, then the parameter's value will be converted
        automatically from ``units`` to :py:attr:`.SpinHamiltonian.units`.

    when_present : str, default "raise error"
        .. versionadded:: 0.4.0

        Action to take if an atom already has a parameter associated with it.
        Case-insensitive. Supported values are:

        - ``"raise error"`` (default): raises an error if an atom already has a parameter
          associated with it.
        - ``"replace"``: replace existing value of the parameter with the new one.
        - ``"add"``: add the value of the parameter to the existing one.
        - ``"mean"``: replace the value of the parameter with the arithmetic mean of
          existing and new parameters.
        - ``"skip"``: Leave existing parameter unchanged and continue without raising an
          error.

    replace : bool, default False
        Whether to replace the value of the parameter if an atom already has a
        parameter associated with it.

        .. deprecated:: 0.4.0
            The ``replace`` argument will be removed in May of 2026. Use
            ``modify="replace"`` instead.

    Raises
    ------

    ValueError
        If an atom already has a parameter associated with it and ``when_present="raise error"``.

    ValueError
        If ``when_present`` has an unsupported value.

    See Also
    --------

    p31
    remove_31
    """

    if replace is not None:
        import warnings

        warnings.warn(
            'The "replace" argument is deprecated since version 0.4.0 and will be removed in May of 2026. Use when_present="replace" instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        if replace:
            when_present = "replace"
        else:
            when_present = "raise error"

    _validate_atom_index(index=alpha, atoms=spinham.atoms)
    spinham._reset_internals()

    parameter = np.array(parameter)

    if units is not None:
        units = _validated_units(units=units, supported_units=_PARAMETER_UNITS)
        parameter = (
            parameter * _PARAMETER_UNITS[units] / _PARAMETER_UNITS[spinham._units]
        )

    # TODO BINARY SEARCH
    # Try to find the place for the new one inside the list
    index = 0
    while index < len(spinham._31):
        # If already present in the model
        if spinham._31[index][0] == alpha:
            # Either replace
            if when_present.lower() == "replace":
                spinham._31[index][1] = parameter
            # Or add
            elif when_present.lower() == "add":
                spinham._31[index][1] = spinham._31[index][1] + parameter
            # Or replace with mean value
            elif when_present.lower() == "mean":
                spinham._31[index][1] = (spinham._31[index][1] + parameter) / 2.0
            # Or do nothing
            elif when_present.lower() == "skip":
                pass
            # Or raise an error
            elif when_present == "raise error":
                raise ValueError(
                    f"(Three spins & one site) parameter is already set for atom {alpha} ('{spinham.atoms.names[alpha]}'."
                )
            else:
                raise ValueError(
                    f'Unsupported value of when_present: "{when_present}". Supported values are: "raise error", "replace", "add", "mean", "skip".'
                )

            return

        # If it should be inserted before current element
        if spinham._31[index][0] > alpha:
            spinham._31.insert(index, [alpha, parameter])
            return

        index += 1

    # If it should be inserted at the end or at the beginning of the list
    spinham._31.append([alpha, parameter])


def _remove_31(spinham, alpha: int) -> None:
    r"""
    Removes a (three spins & one site) parameter from the Hamiltonian.

    Parameters
    ----------

    alpha : int
        Index of an atom, with which the parameter is associated.

        ``0 <= alpha < len(spinham.atoms.names)``.

    See Also
    --------

    p31
    add_31
    """

    _validate_atom_index(index=alpha, atoms=spinham.atoms)

    for i in range(len(spinham._31)):
        # As the list is sorted, there is no point in resuming the search
        # when a larger element is found
        if spinham._31[i][0] > alpha:
            return

        if spinham._31[i][0] == alpha:
            del spinham._31[i]
            spinham._reset_internals()
            return
